from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from biobalm.succession_diagram import SuccessionDiagram

import copy
from biobalm.space_utils import is_subspace
from biobalm.trappist_core import trappist
from biobalm.types import BooleanSpace


def expand_minimal_spaces(
    sd: SuccessionDiagram,
    node_id: int | None,
    size_limit: int | None = None,
    skip_remaining: bool = False,
) -> bool:
    """
    See `SuccessionDiagram.expand_minimal_spaces` for documentation.
    """

    if node_id is None:
        node_id = sd.root()

    pn = sd.node_percolated_petri_net(node_id, compute=True)
    node_space = sd.node_data(node_id)["space"]

    all_minimal_traps = trappist(network=pn, problem="min", ensure_subspace=node_space)
    all_minimal_traps = [(node_space | x) for x in all_minimal_traps]
    # We don't need to duplicate the actual trap spaces, just the list.
    minimal_traps = copy.copy(all_minimal_traps)

    if sd.config["debug"]:
        print(
            f"Start minimal trap SD expansion using {len(minimal_traps)} minimal traps."
        )

    seen = set([node_id])

    stack: list[tuple[int, list[int] | None]] = [(node_id, None)]

    def make_skip_node(
        sd: SuccessionDiagram, node_id: int, all_minimal_traps: list[BooleanSpace]
    ):
        node = sd.node_data(node_id)
        if node["expanded"]:
            return

        skip_edges = 0
        for m_trap in all_minimal_traps:
            if is_subspace(m_trap, sd.node_data(node_id)["space"]):
                m_id = sd._ensure_node(node_id, m_trap)  # type: ignore
                m_data = sd.node_data(m_id)
                m_data["expanded"] = True
                assert sd.node_is_minimal(m_id)
                skip_edges += 1

        node["expanded"] = True
        node["skipped"] = True

        if sd.config["debug"]:
            print(f"[{node_id}] Node skipped with {skip_edges} edges.")

    while len(stack) > 0:
        (node, successors) = stack.pop()
        if successors is None:
            # Only allow successor computation if size limit hasn't been exceeded.
            if (size_limit is not None) and (len(sd) >= size_limit):
                # Size limit reached.
                return False

            successors = sd.node_successors(node, compute=True)
            successors = sorted(successors, reverse=True)  # For determinism!
            # (reversed because we explore the list from the back)

        node_space = sd.node_data(node)["space"]

        # Remove all immediate successors that are already visited or those who
        # do not cover any new minimal trap space.
        while len(successors) > 0:
            if successors[-1] in seen:
                # Everything in seen is expanded, so no need to skip it.
                successors.pop()
                continue
            if len([s for s in minimal_traps if is_subspace(s, node_space)]) == 0:
                skipped = successors.pop()
                if skip_remaining:
                    make_skip_node(sd, skipped, all_minimal_traps)
                continue
            break

        # This node is done and we don't have to push anything onto the stack.
        # In particular, it means that every minimal trap space which is a subset
        # of this node is already in the succession diagram.
        if len(successors) == 0:
            if sd.node_is_minimal(node):
                minimal_traps.remove(sd.node_data(node)["space"])
                if sd.config["debug"]:
                    print(f"Remaining minimal traps: {len(minimal_traps)}.")
            continue

        # At this point, we know that `s` is not visited and it contains
        # at least one minimal trap space that does not appear in the
        # succession diagram yet.

        s = successors.pop()
        seen.add(s)
        # Push the node back with the remaining successors.
        stack.append((node, successors))
        # Push the successor onto the stack.
        stack.append((s, None))

        if sd.config["debug"]:
            print(f"[{s}] Expanding...")

    assert len(minimal_traps) == 0
    return True



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, cast

if TYPE_CHECKING:
    from typing import Iterator

import copy
import networkx as nx  # type: ignore
from biodivine_aeon import AsynchronousGraph, BooleanNetwork, VertexSet

# Attractor detection algorithms.
from biobalm._sd_attractors.attractor_candidates import compute_attractor_candidates
from biobalm._sd_attractors.attractor_symbolic import (
    compute_attractors_symbolic,
    symbolic_attractor_fallback,
)

# SD expansion algorithms/heuristics.
from biobalm._sd_algorithms.expand_attractor_seeds import expand_attractor_seeds
from biobalm._sd_algorithms.expand_bfs import expand_bfs
from biobalm._sd_algorithms.expand_dfs import expand_dfs
from biobalm._sd_algorithms.expand_minimal_spaces import expand_minimal_spaces
from biobalm._sd_algorithms.expand_source_SCCs import expand_source_SCCs
from biobalm._sd_algorithms.expand_source_blocks import expand_source_blocks
from biobalm._sd_algorithms.expand_to_target import expand_to_target

from biobalm.interaction_graph_utils import (
    cleanup_network,
    feedback_vertex_set,
    source_SCCs,
)
from biobalm.petri_net_translation import (
    extract_source_variables,
    network_to_petrinet,
    restrict_petrinet_to_subspace,
)
from biobalm.space_utils import (
    percolate_network,
    percolate_space,
    space_unique_key,
    is_subspace,
)
from biobalm.trappist_core import trappist
from biobalm.types import (
    BooleanSpace,
    NodeData,
    SuccessionDiagramState,
    SuccessionDiagramConfiguration,
)


class SuccessionDiagram:
    """
    Succession diagram of a Boolean network.

    This encodes relationships between minimal trap spaces and can be used for
    attractor detection and control. Note that the succession diagram is
    expanded lazily, so it is not built until `build` or a similar method is
    called.


    Examples
    --------
    >>> import biobalm
    >>> sd = biobalm.SuccessionDiagram.from_rules(\"""
    ...     A, B
    ...     B, A & C
    ...     C, !A | B
    ... \""")
    >>> print(sd.summary()) # not built yet!
    Succession Diagram with 1 nodes and depth 0.
    State order: A, B, C
    <BLANKLINE>
    Attractors in diagram:
    <BLANKLINE>
    >>> sd.build()
    >>> print(sd.summary()) # now it's built
    Succession Diagram with 3 nodes and depth 1.
    State order: A, B, C
    <BLANKLINE>
    Attractors in diagram:
    <BLANKLINE>
    minimal trap space 001
    ...................001
    <BLANKLINE>
    minimal trap space 111
    ...................111
    <BLANKLINE>
    """

    __slots__ = (
        "network",
        "symbolic",
        "petri_net",
        "nfvs",
        "dag",
        "node_indices",
        "config",
    )

    def __init__(
        self,
        network: BooleanNetwork,
        config: SuccessionDiagramConfiguration | None = None,
    ):
        if config is None:
            config = SuccessionDiagram.default_config()
        self.config = config

        # Original Boolean network.
        self.network: BooleanNetwork = cleanup_network(network)
        """
        The Boolean network represented as a `biodivine_aeon.BooleanNetwork` object.
        """

        self.symbolic: AsynchronousGraph = AsynchronousGraph(self.network)
        """
        The symbolic representation of the network using `biodivine_aeon.AsynchronousGraph`.
        """

        self.petri_net: nx.DiGraph = network_to_petrinet(network)
        """
        The Petri net representation of the network (see :mod:`petri_net_translation<biobalm.petri_net_translation>`).
        """

        if self.config["debug"]:
            print(
                f"Generated global Petri net with {len(self.petri_net.nodes)} nodes and {len(self.petri_net.edges)} edges."
            )

        self.nfvs: list[str] | None = None
        """
        The negative feedback vertex set of `SuccessionDiagram.network`, or `None` if not computed yet.
        """

        self.dag: nx.DiGraph = nx.DiGraph()
        """
        The directed acyclic graph (DAG) representation of the succession
        diagram structure.
        """

        self.node_indices: dict[int, int] = {}
        """
        A dictionary mapping subspace keys to their positions in the succession
        diagram (see :func:`biobalm.space_utils.space_unique_key`).
        """

        # Create an un-expanded root node.
        self._ensure_node(None, {})

    def __getstate__(self) -> SuccessionDiagramState:
        return {
            "network_rules": self.network.to_aeon(),
            "petri_net": self.petri_net,
            "nfvs": self.nfvs,
            "dag": self.dag,
            "node_indices": self.node_indices,
            "config": self.config,
        }

    def __setstate__(self, state: SuccessionDiagramState):
        # In theory, the network should be cleaned-up at this point, but just in case...
        self.network = cleanup_network(BooleanNetwork.from_aeon(state["network_rules"]))
        self.symbolic = AsynchronousGraph(self.network)
        self.petri_net = state["petri_net"]
        self.nfvs = state["nfvs"]
        self.dag = state["dag"]
        self.node_indices = state["node_indices"]
        self.config = state["config"]

    def __len__(self) -> int:
        """
        Returns the number of nodes in this `SuccessionDiagram`.
        """
        return self.dag.number_of_nodes()

    @staticmethod
    def default_config() -> SuccessionDiagramConfiguration:
        return {
            "debug": False,
            "max_motifs_per_node": 100_000,
            "nfvs_size_threshold": 2_000,
            "pint_goal_size_limit": 8_192,
            "attractor_candidates_limit": 100_000,
            "retained_set_optimization_threshold": 1_000,
            "minimum_simulation_budget": 1_000,
        }

    @staticmethod
    def from_rules(
        rules: str,
        format: Literal["bnet", "aeon", "sbml"] = "bnet",
        config: SuccessionDiagramConfiguration | None = None,
    ) -> SuccessionDiagram:
        """
        Generate a succession diagram from the given string.

        Parameters
        ----------
        rules : str
            The string representation of the network rules.
        format : Literal['bnet', 'aeon', 'sbml']
            The format of the string. One of `"bnet"`, `"aeon"`, or `"sbml"`.
            Defaults to `"bnet"`.
        config : SuccessionDiagramConfiguration | None
            An optional configuration object with internal settings
            and default values.

        Returns
        -------
        SuccessionDiagram
            The generated succession diagram. Initially unexpanded.
        """

        if format == "bnet":
            return SuccessionDiagram(BooleanNetwork.from_bnet(rules), config)
        elif format == "aeon":
            return SuccessionDiagram(BooleanNetwork.from_aeon(rules), config)
        elif format == "sbml":
            return SuccessionDiagram(BooleanNetwork.from_sbml(rules), config)
        else:
            raise ValueError(f"Unknown format: {format}")

    @staticmethod
    def from_file(
        path: str, config: SuccessionDiagramConfiguration | None = None
    ) -> SuccessionDiagram:
        """
        Read a `BooleanNetwork` from the given file path. The format is automatically inferred from
        the file extension.

        Optionally, you can also supply a configuration object to customize the
        resulting succession diagram.
        """
        return SuccessionDiagram(BooleanNetwork.from_file(path), config)

    def expanded_attractor_candidates(self) -> dict[int, list[BooleanSpace]]:
        """
        Attractor candidates for each expanded node. The candidate list is
        computed for nodes where it is not known yet.

        Collectively, for every attractor in every expanded node, this returns
        at least one candidate state for said attractor. However, for complex
        attractors, there can be more than one candidate state, and there can
        be candidates that are not contained in any attractor.

        *If called before the `SuccessionDiagram` is fully built, this will not
        be a complete accounting of attractors, since any node that isn't expanded
        is not included in the result.*

        See also:
         - :meth:`expanded_attractor_seeds<SuccessionDiagram.expanded_attractor_seeds>`
         - :meth:`expanded_attractor_seeds<SuccessionDiagram.expanded_attractor_sets>`
         - :meth:`expanded_attractor_seeds<SuccessionDiagram.node_attractor_candidates>`
         - :meth:`expanded_attractor_seeds<SuccessionDiagram.node_attractor_seeds>`
         - :meth:`expanded_attractor_seeds<SuccessionDiagram.node_attractor_sets>`

        Returns
        -------
        dict[int,list[BooleanSpace]]
            Each key is the id of an expanded succession diagram node, whose
            corresponding value is a list of attractor candidates for that node. Note
            that one succession diagram node can have multiple attractors. Ordering
            of the lists in the returned dictionary is not guaranteed.

        Example
        -------
        >>> import biobalm
        >>> sd = biobalm.SuccessionDiagram.from_rules(\"""
        ...     A, B
        ...     B, A & C
        ...     C, !A | B
        ... \""")
        >>> sd.build()
        >>> eas = sd.expanded_attractor_candidates()
        >>> for id, atts in sorted(eas.items()):
        ...     for x in atts:
        ...         print(f"{id}: {dict(sorted(x.items()))}")
        1: {'A': 0, 'B': 0, 'C': 1}
        2: {'A': 1, 'B': 1, 'C': 1}
        """
        res: dict[int, list[BooleanSpace]] = {}
        for id in self.expanded_ids():
            atts = self.node_attractor_candidates(id, compute=True)
            if not atts:  # no attractors for this node
                continue
            res[id] = atts

        return res

    def expanded_attractor_seeds(self) -> dict[int, list[BooleanSpace]]:
        """
        Attractor seeds for each expanded node. The seed list is
        computed for nodes where it is not known yet.

        Collectively, for every attractor in every expanded node, this returns
        exactly one seed state for said attractor.

        *If called before the `SuccessionDiagram` is fully built, this will not
        be a complete accounting of attractors, since any node that isn't expanded
        is not included in the result.*

        See also:
         - :meth:`expanded_attractor_seeds<SuccessionDiagram.expanded_attractor_candidates>`
         - :meth:`expanded_attractor_seeds<SuccessionDiagram.expanded_attractor_sets>`
         - :meth:`expanded_attractor_seeds<SuccessionDiagram.node_attractor_candidates>`
         - :meth:`expanded_attractor_seeds<SuccessionDiagram.node_attractor_seeds>`
         - :meth:`expanded_attractor_seeds<SuccessionDiagram.node_attractor_sets>`

        Returns
        -------
        dict[int,list[BooleanSpace]]
            Each key is the id of an expanded succession diagram node, whose
            corresponding value is a list of attractor seeds for that node. Note
            that one succession diagram node can have multiple attractors. Ordering
            of the lists in the returned dictionary is not guaranteed.

        Example
        -------

        *Note that for this simple network, attractor candidates and attractor
        seeds are the same states.*

        >>> import biobalm
        >>> sd = biobalm.SuccessionDiagram.from_rules(\"""
        ...     A, B
        ...     B, A & C
        ...     C, !A | B
        ... \""")
        >>> sd.build()
        >>> eas = sd.expanded_attractor_seeds()
        >>> for id, atts in sorted(eas.items()):
        ...     for x in atts:
        ...         print(f"{id}: {dict(sorted(x.items()))}")
        1: {'A': 0, 'B': 0, 'C': 1}
        2: {'A': 1, 'B': 1, 'C': 1}
        """
        res: dict[int, list[BooleanSpace]] = {}
        for id in self.expanded_ids():
            atts = self.node_attractor_seeds(id, compute=True)
            if not atts:  # no attractors for this node
                continue
            res[id] = atts

        return res

    def expanded_attractor_sets(self) -> dict[int, list[VertexSet]]:
        """
        Attractor sets for each expanded node. The sets are
        computed for nodes where they are not known yet.

        These sets represent the complete collection of attractor states in each
        expanded node. For fixed-point attractors, this is effectively equivalent
        to the attractor seeds. For complex attractors, this set contains the
        full attractor. Hence, it is harder to compute but can facilitate richer
        post-processing and analysis.

        *If called before the `SuccessionDiagram` is fully built, this will not
        be a complete accounting of attractors, since any node that isn't expanded
        is not included in the result.*

        See also:
         - :meth:`expanded_attractor_seeds<SuccessionDiagram.expanded_attractor_candidates>`
         - :meth:`expanded_attractor_seeds<SuccessionDiagram.expanded_attractor_seeds>`
         - :meth:`expanded_attractor_seeds<SuccessionDiagram.node_attractor_candidates>`
         - :meth:`expanded_attractor_seeds<SuccessionDiagram.node_attractor_seeds>`
         - :meth:`expanded_attractor_seeds<SuccessionDiagram.node_attractor_sets>`

        Returns
        -------
        dict[int,list[biodivine_aeon.VertexSet]]
            Each key is the id of an expanded succession diagram node, whose
            corresponding value is a list of attractor sets for that node. Note
            that one succession diagram node can have multiple attractors. Ordering
            of the lists in the returned dictionary is not guaranteed.

        Example
        -------
        >>> import biobalm
        >>> sd = biobalm.SuccessionDiagram.from_rules(\"""
        ...     A, B
        ...     B, A & C
        ...     C, !A | B
        ... \""")
        >>> sd.build()
        >>> eas = sd.expanded_attractor_sets()
        >>> for id, atts in sorted(eas.items()):
        ...     for x in atts:
        ...         print(f"{id}: {x}")
        1: VertexSet(cardinality=1, symbolic_size=5)
        2: VertexSet(cardinality=1, symbolic_size=5)
        """
        res: dict[int, list[VertexSet]] = {}
        for id in self.expanded_ids():
            atts = self.node_attractor_sets(id, compute=True)
            if not atts:  # no attractors for this node
                continue
            res[id] = atts

        return res

    def summary(self) -> str:
        """
        Return a summary of the succession diagram as a string.
        """
        var_ordering = sorted(
            [self.network.get_variable_name(v) for v in self.network.variables()]
        )
        report_string = (
            f"Succession Diagram with {len(self)} nodes and depth {self.depth()}.\n"
            f"State order: {', '.join(var_ordering)}\n\n"
            "Attractors in diagram:\n\n"
        )
        for node in self.node_ids():
            try:
                attrs = self.node_attractor_seeds(node, compute=False)
            except KeyError:
                continue

            if not attrs:
                continue

            space = self.node_data(node)["space"]

            if self.node_is_minimal(node):
                space_str_prefix = "minimal trap space "
            else:
                space_str_prefix = "motif avoidance in "
            space_str = ""
            for var in var_ordering:
                if var in space:
                    space_str += str(space[var])
                else:
                    space_str += "*"
            report_string += f"{space_str_prefix}{space_str}\n"
            for attr in attrs:
                attr_str = "".join(str(v) for _, v in sorted(attr.items()))
                report_string += "." * len(space_str_prefix) + f"{attr_str}\n"
            report_string += "\n"
        # remove final extra newline and return
        return report_string[:-1]

    def root(self) -> int:
        """
        Return the ID of the root node.
        """
        return 0

    def depth(self) -> int:
        """
        Compute the maximal node depth in the diagram.

        Depth is counted from zero (root has depth zero).
        """
        d = 0
        for node in cast(set[int], self.dag.nodes()):
            d = max(d, self.node_data(int(node))["depth"])
        return d

    def node_ids(self) -> Iterator[int]:
        """
        Iterator over all available node IDs.
        """
        for i in range(len(self)):
            yield i

    def stub_ids(self) -> Iterator[int]:
        """
        Iterator over all node IDs that are currently *not* expanded.
        """
        for i in range(len(self)):
            if not self.node_data(i)["expanded"]:
                yield i

    def expanded_ids(self) -> Iterator[int]:
        """
        Iterator over all node IDs that are currently expanded.
        """
        for i in range(len(self)):
            if self.node_data(i)["expanded"]:
                yield i

    def minimal_trap_spaces(self) -> list[int]:
        """
        List of node IDs that represent the minimal trap spaces within this
        succession diagram.

        Note that stub nodes do not count as minimal!
        """
        return [i for i in self.expanded_ids() if self.node_is_minimal(i)]

    def find_node(self, node_space: BooleanSpace) -> int | None:
        """
        Return the ID of the node matching the provided `node_space`, or `None`
        if no such node exists in this succession diagram.

        Parameters
        ----------
        node_space: BooleanSpace
            The space of the node to find.

        Returns
        -------
        int | None
            The ID of the node matching the provided `node_space`, or `None`
            if no such node exists in this succession diagram.
        """
        try:
            key = space_unique_key(node_space, self.network)  # throws IndexError
            if key in self.node_indices:
                return self.node_indices[key]
            else:
                return None
        except IndexError:
            # If `space_unique_key` finds a variable that does not exist in this
            # `SuccessionDiagram`, it throws an `IndexError`. This can happen
            # for example if we are comparing two succession diagrams based on
            # completely different networks.
            return None

    def is_subgraph(self, other: SuccessionDiagram) -> bool:
        """
        Returns `True` if this succession diagram is a subgraph of the `other`
        succession diagram.

        Note that this function works even for diagrams based on different
        Boolean networks, as long as both succession diagrams only depend on the
        same subset of network variables.

        WARNING: This does not take into account the stable motifs on individual
        edges. Just the subspaces associated with nodes and the presence of
        edges between nodes.

        Parameters
        ----------
        other: SuccessionDiagram
            The other succession diagram.

        Returns
        -------
        bool
            `True` if this succession diagram is a subgraph of the `other`
            succession diagram.
        """
        # Every stub node is reachable through an expanded node and
        # thus will be checked by the following code.
        for i in self.expanded_ids():
            other_i = other.find_node(self.node_data(i)["space"])
            if other_i is None:
                return False
            my_successors = self.node_successors(i)
            other_successors = []
            if other.node_data(other_i)["expanded"]:
                other_successors = other.node_successors(other_i)

            for my_s in my_successors:
                other_s = other.find_node(self.node_data(my_s)["space"])
                if other_s not in other_successors:
                    return False
        return True

    def is_isomorphic(self, other: SuccessionDiagram) -> bool:
        """
        Returns `True` if the two succession diagrams are isomorphic.

        Note that this function works even for diagrams based on different
        Boolean networks, as long as both succession diagrams only depend on the
        same subset of network variables.

        WARNING: This does not take into account the stable motifs on individual
        edges. Just the subspaces associated with nodes and the presence of
        edges between nodes.

        Parameters
        ----------
        other: SuccessionDiagram
            The other succession diagram.

        Returns
        -------
        bool
            `True` if the two succession diagrams are isomorphic.
        """
        return self.is_subgraph(other) and other.is_subgraph(self)

    def node_data(self, node_id: int) -> NodeData:
        """
        Get the data associated with the provided `node_id`.

        Returns a `NodeData` object with the following attributes:

        - `depth`: The depth of the node.
        - `space`: The sub-space of the node.
        - `expanded`: Whether the node is expanded or not.
        - `percolated_network`: [`None` if not computed] The percolation of
        `SuccessionDiagram.network` with respect to the node's `space`.
        - `percolated_petri_net`: [`None` if not computed] The percolation of
        `SuccessionDiagram.petri_net` with respect to the node's `space`.
        - `percolated_nfvs`: [`None` if not computed] The NFVS of `percolated_network`.
        - `attractor_candidates`: [`None` if not computed] A collection of states
        that collectively cover every attractor of this node.
        - `attractor_seeds`: [`None` if not computed] A collection of states
        one-to-one corresponding to the attractors of this node.
        - `attractor_sets`: [`None` if not computed] A complete collection of
        attractors of this node.

        See :class:`biobalm.types.NodeData` for more information.

        Parameters
        ----------
        node_id: int
            The ID of the node.

        Returns
        -------
        NodeData
            The data associated with the provided `node_id`. Note that at
            runtime, this object is an untyped dictionary.
        """
        return cast(NodeData, self.dag.nodes[node_id])

    def reclaim_node_data(self):
        """
        Removes non-essential data from the `NodeData` dictionary of each node.

        This method can be used to reduce the memory footprint of the succession
        diagram, especially before serialization. However, note that this can
        also slow down subsequent computations if the erased data needs
        to be re-computed.

        The method removes the `percolated_network`, `percolated_petri_net`,
        and the `percolated_nfvs`. Furthermore, if `attractor_seeds` are known,
        it erases the `attractor_candidates`, since seeds can be used for the
        same tasks.
        """

        for node_id in self.node_ids():
            data = self.node_data(node_id)
            data["percolated_network"] = None
            data["percolated_petri_net"] = None
            data["percolated_nfvs"] = None
            if data["attractor_seeds"] is not None:
                data["attractor_candidates"] = None

    def node_is_minimal(self, node_id: int) -> bool:
        """
        True if the node represents a minimal trap space.

        Parameters
        ----------
        node_id: int
            The ID of the node.

        Returns
        -------
        bool
            `True` if the node is expanded and it has no successors, i.e. it is a
            minimal trap space.
        """
        is_leaf: bool = self.dag.out_degree(node_id) == 0  # type: ignore
        return is_leaf and self.dag.nodes[node_id]["expanded"]  # type: ignore

    def node_successors(self, node_id: int, compute: bool = False) -> list[int]:
        """
        Return the successor nodes for the given `node_id`.

        If the node is already expanded, known results are simply returned. If
        the node is not expanded, but `compute` is set to `True`, then the node
        is expanded and the newly computed results are returned. If the node is
        not expanded and `compute` is set to `False`, the method raises a
        `KeyError` exception.

        The default behaviour intentionally does not compute successors to
        prevent "accidental complexity".

        WARNING: We do not guarantee the order of the returned nodes. If you
        need the successors in a repeatable order, you should sort the list
        manually.

        Also note that if the given `node_id` already has associated attractor
        data but is not expanded, this data will be deleted as it is no longer
        up to date.

        Parameters
        ----------
        node_id: int
            The ID of the node.
        compute: bool
            Whether to compute the successors if they are not already known.

        Returns
        -------
        list[int]
            The list of successor node ids.
        """
        node = self.node_data(node_id)
        if not node["expanded"] and not compute:
            raise KeyError(f"Node {node_id} is not expanded.")
        if not node["expanded"]:
            self._expand_one_node(node_id)

        return list(self.dag.successors(node_id))  # type: ignore

    def node_attractor_candidates(
        self,
        node_id: int,
        compute: bool = False,
        greedy_asp_minification: bool = True,
        simulation_minification: bool = True,
        pint_minification: bool = False,
    ) -> list[BooleanSpace]:
        """
        Return the list of attractor candidate states for the given `node_id`.

        If attractor candidates are not computed but seeds are, returns
        attractor seeds, as these are also valid as candidate states, but
        even more precise.

        Similar to :meth:`node_successors`, the method either computes the
        data if unknown, or throws an exception, depending on the `compute`
        flag. If `compute` is set to `True`, additional flags can be used
        to adjust the candidate identification process (see *Parameters*).

        Note that you can compute attractor candidates for nodes that are not expanded,
        but (a) multiple unexpanded nodes can contain the same attractor, and hence also
        the same/similar candidates (i.e. you can "discover" the same attractor in multiple
        unexpanded nodes, if the nodes intersect), and (b) this data is erased if the
        node is later expanded.

        Parameters
        ----------
        node_id: int
            The ID of the node.
        compute: bool
            Whether to compute the attractor candidates if they are not already known.
        greedy_asp_minification: bool
            Indicate that the initial candidate set should be first greedily minified
            through repeated ASP queries. [Default: True]
        simulation_minification: bool
            Indicate that the candidate set should be refined through stochastic
            simulation. [Default: True]
        pint_minification: bool
            Indicate that the candidate set should be refined through reachability
            analysis using `pint`. Only enable this option if you actually have `pint`
            installed (it is an optional dependency). [Default: False]

        Returns
        -------
        list[BooleanSpace]
            The list of attractor candidate states.
        """
        node = self.node_data(node_id)

        candidates = node["attractor_candidates"]

        # If candidates are already cleared, but seeds are known, we can
        # just use seeds.
        if candidates is None and node["attractor_seeds"] is not None:
            return node["attractor_seeds"]

        if candidates is None and not compute:
            raise KeyError(f"Attractor candidates not computed for node {node_id}.")

        if candidates is None:
            candidates = compute_attractor_candidates(
                self,
                node_id,
                greedy_asp_minification,
                simulation_minification,
                pint_minification,
            )
            node["attractor_candidates"] = candidates

            # If the computed candidates are actually valid as seeds, just
            # propagate the value so that it doesn't need to be computed later.
            node_is_pseudo_minimal = (not node["expanded"]) or self.node_is_minimal(
                node_id
            )
            if len(candidates) == 0 or (
                node_is_pseudo_minimal and len(candidates) == 1
            ):
                node["attractor_seeds"] = candidates

        return candidates

    def node_attractor_seeds(
        self,
        node_id: int,
        compute: bool = False,
        symbolic_fallback: bool = False,
    ) -> list[BooleanSpace]:
        """
        Return the list of attractor seed states for the given `node_id`.

        Similar to :meth:`node_successors`, the method either computes the
        data if unknown, or throws an exception, depending on the `compute`
        flag.

        Note that the same considerations regarding attractors in unexpanded
        nodes apply as for :meth:`node_attractor_candidates`.

        Parameters
        ----------
        node_id: int
            The ID of the node.
        compute: bool
            Whether to compute the attractor seeds if they are not already known.
        symbolic_fallback: bool
            If active, the method will attempt to compute the attractor seeds fully
            symbolically if the default NFVS-based method fails. However, note that
            the program can become unresponsive if the symbolic encoding of the
            result grows to be too large. [Default: False]

        Returns
        -------
        list[BooleanSpace]
            The list of attractor seed states.
        """
        node = self.node_data(node_id)

        seeds = node["attractor_seeds"]

        if seeds is None and not compute:
            raise KeyError(f"Attractor seeds not computed for node {node_id}.")

        try:
            if seeds is None:
                candidates = self.node_attractor_candidates(node_id, compute=True)
                # Typically, this should be done when computing the candidates, but just in case
                # something illegal happended... if we can show that the current candidate set
                # is optimal, we just keep it and don't compute the attractors symbolically.
                node_is_pseudo_minimal = (not node["expanded"]) or self.node_is_minimal(
                    node_id
                )
                if len(candidates) == 0 or (
                    node_is_pseudo_minimal and len(candidates) == 1
                ):
                    node["attractor_seeds"] = candidates
                    seeds = candidates
                else:
                    result = compute_attractors_symbolic(
                        self, node_id, candidate_states=candidates, seeds_only=True
                    )
                    node["attractor_seeds"] = result[0]
                    # At this point, attractor_sets could be `None`, but that
                    # is valid, as long as we actually compute them later when
                    # they are needed.
                    node["attractor_sets"] = result[1]
                    seeds = result[0]

                # Release memory once attractor seeds are known. We might need these
                # for attractor set computation later, but only if the seeds are not empty.
                if len(seeds) == 0:
                    node["percolated_network"] = None
                    node["percolated_nfvs"] = None
                    node["percolated_petri_net"] = None
        except RuntimeError as e:
            if not symbolic_fallback:
                raise e

            # The NFVS method failed, likely because the candidate set was too large.
            # We can still try to fix this and compute the attractors symbolically.
            (seeds, sets) = symbolic_attractor_fallback(self, node_id)

            node["attractor_seeds"] = seeds
            node["attractor_sets"] = sets

            if len(seeds) == 0:
                node["percolated_network"] = None
                node["percolated_nfvs"] = None
                node["percolated_petri_net"] = None

        return seeds

    def node_attractor_sets(
        self,
        node_id: int,
        compute: bool = False,
    ) -> list[VertexSet]:
        """
        Return the list of attractor sets for the given `node_id`.

        Similar to :meth:`node_successors`, the method either computes the
        data if unknown, or throws an exception, depending on the `compute`
        flag.

        Note that the same considerations regarding attractors in unexpanded
        nodes apply as for :meth:`node_attractor_candidates`.

        Parameters
        ----------
        node_id: int
            The ID of the node.
        compute: bool
            Whether to compute the attractor sets if they are not already known.

        Returns
        -------
        list[biodivine_aeon.VertexSet]
            The list of attractor sets.
        """
        node = self.node_data(node_id)

        sets = node["attractor_sets"]

        if sets is None and not compute:
            raise KeyError(f"Attractor sets not computed for node {node_id}.")

        if sets is None:
            seeds = self.node_attractor_seeds(node_id, compute=True)
            result: tuple[list[BooleanSpace], list[VertexSet] | None] = ([], [])
            if len(seeds) > 0:
                result = compute_attractors_symbolic(
                    self, node_id, candidate_states=seeds
                )
            assert result[1] is not None
            node["attractor_sets"] = result[1]
            sets = result[1]

        return sets

    def node_percolated_nfvs(self, node_id: int, compute: bool = False) -> list[str]:
        """
        Approximate minimum negative feedback vertex set on the Boolean network
        percolated to the node's sub-space.

        Similar to :meth:`node_successors`, the method either computes the
        data if unknown, or throws an exception, depending on the `compute`
        flag.

        See :func:`biobalm.interaction_graph_utils.feedback_vertex_set` for
        further details.

        Parameters
        ----------
        node_id: int
            The ID of the node.
        compute: bool
            Whether to compute the node NFVS if it is not already known.

        Returns
        -------
        list[str]
            The negative feedback vertex set, as a list of node names.
        """

        assert node_id in self.dag.nodes

        node = self.node_data(node_id)

        if node["percolated_nfvs"] is None and not compute:
            raise KeyError(f"NFVS not computed for node {node_id}.")

        if node["percolated_nfvs"] is None:
            percolated_network = self.node_percolated_network(node_id, compute)
            percolated_size = percolated_network.variable_count()
            if percolated_size < self.config["nfvs_size_threshold"]:
                # Computing the *negative* variant of the FVS is surprisingly costly.
                # Hence it mostly makes sense for the smaller networks only.
                nfvs = feedback_vertex_set(percolated_network, parity="negative")
            else:
                nfvs = feedback_vertex_set(percolated_network)
            node["percolated_nfvs"] = nfvs
        else:
            nfvs = node["percolated_nfvs"]

        return nfvs

    def node_percolated_network(
        self, node_id: int, compute: bool = False
    ) -> BooleanNetwork:
        """
        The Boolean network percolated to the node's sub-space, with
        constant variables removed.

        Similar to :meth:`node_successors`, the method either computes the
        data if unknown, or throws an exception, depending on the `compute`
        flag.

        Parameters
        ----------
        node_id: int
            The ID of the node.
        compute: bool
            Whether to compute the node NFVS if it is not already known.

        Returns
        -------
        biodivine_aeon.BooleanNetwork
            The percolated Boolean network.
        """

        assert node_id in self.dag.nodes

        node = self.node_data(node_id)
        network = node["percolated_network"]

        node_space = node["space"]

        if len(node_space) == self.network.variable_count():
            # If fixed point, no need to compute, the network is always empty.
            return BooleanNetwork()

        if network is None and not compute:
            raise KeyError(f"Percolated network not computed for node {node_id}.")

        if network is None:
            network = percolate_network(
                self.network, node_space, self.symbolic, remove_constants=True
            )
            if self.config["debug"]:
                print(
                    f"[{node_id}] Computed percolated network with {network.variable_count()} variables (vs {self.network.variable_count()})."
                )
            node["percolated_network"] = network

        return network

    def node_percolated_petri_net(
        self,
        node_id: int,
        compute: bool = False,
        parent_id: int | None = None,
    ) -> nx.DiGraph:
        """
        The Petri net representation of the Boolean network percolated to the
        node's sub-space (with constant variables removed).

        Similar to :meth:`node_successors`, the method either computes the
        data if unknown, or throws an exception, depending on the `compute`
        flag.

        Parameters
        ----------
        node_id: int
            The ID of the node.
        compute: bool
            Whether to compute the node NFVS if it is not already known.
        parent_id: int | None
            If given, the percolation process starts with the Petri net of the given
            parent node (if computed). If parent is not given, the percolation starts
            with `SuccessionDiagram.petri_net`, which can be slower but yields the
            same result.

        Returns
        -------
        networkx.DiGraph
            The percolated Boolean network.
        """

        assert node_id in self.dag.nodes

        node = self.node_data(node_id)
        percolated_pn = node["percolated_petri_net"]

        node_space = node["space"]

        if len(node_space) == self.network.variable_count():
            # If fixed point, the result is always empty.
            return nx.DiGraph()

        if percolated_pn is None and not compute:
            raise KeyError(f"Percolated network not computed for node {node_id}.")

        if parent_id is None:
            # If no parent node is directly provided, we can try to use a cached one.
            parent_id = node["parent_node"]

        if percolated_pn is None:
            base_pn = self.petri_net
            percolate_space = node_space
            if parent_id is not None:
                parent_pn = self.node_data(parent_id)["percolated_petri_net"]
                if parent_pn is not None:
                    base_pn = parent_pn
                    percolate_space = node_space

            percolated_pn = restrict_petrinet_to_subspace(base_pn, percolate_space)

            if self.config["debug"]:
                print(
                    f"[{node_id}] Generated Petri net restriction with {len(percolated_pn.nodes)} nodes and {len(percolated_pn.edges)} edges."
                )

            node["percolated_petri_net"] = percolated_pn

        return percolated_pn

    def edge_all_stable_motifs(
        self, parent_id: int, child_id: int, reduced: bool = False
    ) -> list[BooleanSpace]:
        """
        Similar to `edge_stable_motif`, but returns all motifs associated with an edge.
        """
        all_motifs: list[BooleanSpace] = cast(list[BooleanSpace], self.dag.edges[parent_id, child_id]["all_motifs"])  # type: ignore
        if reduced:
            result: list[BooleanSpace] = []
            node_space = self.node_data(parent_id)["space"]
            for m in all_motifs:
                result.append({k: v for k, v in m.items() if k not in node_space})
            return result
        else:
            return all_motifs

    def edge_stable_motif(
        self, parent_id: int, child_id: int, reduced: bool = False
    ) -> BooleanSpace:
        """
        Return the stable motif for the given parent-child edge.

        If `reduced` is set to `False` (default), the unpercolated stable
        motif trap space corresponding to the child node is returned; this
        includes the nodes that are fixed in the percolated trap space of the
        parent node. If `reduced` is set to `True`, the nodes fixed in the
        parent are removed (and thus the reduced stable motif is not a trap
        space of the original network, but is a maximal trap space in the
        network reduced by the parent node).

        Parameters
        ----------
        parent_id: int
            The ID of the parent node.
        child_id: int
            The ID of the child node.
        reduced: bool
            Whether to return the reduced stable motif.

        Returns
        -------
        BooleanSpace
            The stable motif (maximal trap space) represented by the edge.
        """

        if reduced:
            return cast(
                BooleanSpace,
                {
                    k: v
                    for k, v in self.dag.edges[parent_id, child_id]["motif"].items()  # type: ignore
                    if k not in self.node_data(parent_id)["space"]
                },
            )
        else:
            return cast(BooleanSpace, self.dag.edges[parent_id, child_id]["motif"])

    def component_subdiagram(
        self,
        component_variables: list[str],
        node_id: int | None = None,
    ) -> SuccessionDiagram:
        """
        Return an *unexpanded* `SuccessionDiagram` that is restricted to
        a subnetwork induced by the provided `component_variables`.  Furthermore,
        If `node_id` is given, the subnetwork is first percolated to the
        subspace of the specified node.

        The `component_variables` must be backward-closed in the considered network
        (i.e. either the full network, or the percolated network if `node_id` is given),
        meaning there is no variable outside this list that regulates any variable in the
        subnetwork. If this is not satisfied, the function will fail while
        creating the subnetwork.

        Also note that the symbolic encoding of the new network is not
        compatible with the encoding of the original network, because the
        underlying networks have different sets of variables.

        Parameters
        ----------
        component_variables : list[str]
            Names of variables which induce the subnetwork of the resulting
            succession diagram.
        node_id : int | None
            The ID of a succession diagram node that will define a subspace
            to which the subnetwork is percolated. If not given, the full
            network is considered.

        Returns
        -------
        SuccessionDiagram
            An unexpanded succession diagram of the subnetwork.
        """

        network = self.network
        if node_id is not None:
            network = self.node_percolated_network(node_id, compute=True)

        to_remove = [
            v for v in network.variable_names() if v not in component_variables
        ]
        component_bn = network.drop(to_remove)
        config_copy: SuccessionDiagramConfiguration = copy.copy(self.config)
        return SuccessionDiagram(component_bn, config_copy)

    def source_scc_subdiagrams(
        self,
        node_id: int | None = None,
    ) -> Iterator[SuccessionDiagram]:
        """
        Return unexpanded subdiagrams for the source SCCs in a node subspace.

        Note that the symbolic encoding of the new network is not
        compatible with the encoding of the original network, because the
        underlying networks have different sets of variables.

        Parameters
        ----------
        node_id : int | None
            The ID of a succession diagram node that will define a subspace on
            which the subnetworks should be considered. By default, the root node
            is used.

        Returns
        -------
        Iterator[SuccessionDiagram]
            An iterator over unexpanded succession diagrams of the subnetwork.
        """

        if node_id is None:
            node_id = self.root()

        reference_bn = self.node_percolated_network(node_id, compute=True)
        source_scc_list = source_SCCs(reference_bn)

        for component_variables in source_scc_list:
            yield self.component_subdiagram(component_variables, node_id)

    def build(self):
        """
        Expand the succession diagram and search for attractors using default methods.
        """
        self.expand_block()
        for node_id in self.node_ids():
            self.node_attractor_seeds(node_id, compute=True)

    def expand_scc(self, find_motif_avoidant_attractors: bool = True) -> bool:
        """
        Expand the succession diagram using the source SCC method.
        """
        return expand_source_SCCs(self, check_maa=find_motif_avoidant_attractors)

    def expand_block(
        self,
        find_motif_avoidant_attractors: bool = True,
        size_limit: int | None = None,
        optimize_source_nodes: bool = True,
        exact_attractor_detection: bool = False,
    ) -> bool:
        """
        Expand the succession diagram using the source block method.

        There is a minor difference in behavior depending on `find_motif_avoidant_attractors`.
        If set to `False`, the expansion only expands one "source block" for each node,
        without checking any attractor properties. If set to `True`, the expansion might
        expand some nodes fully to uncover nodes that precisely cover motif
        avoidant attractors. As a byproduct, if set to `True` and no motif avoidant attractors
        are detected for some node, this is result is saved and the attractors don't
        need to be recomputed later.

        By default, the method also detects any source nodes and directly expands these
        into trap spaces where all source nodes are fixed. This has no correctness impact on
        attractor search and always produces a smaller succession diagram, but if you need to
        obtain a succession diagram where this does not happen (e.g. for testing), you can turn
        this off using `optimize_source_nodes`.

        If `exact_attractor_detection` is selected, the method will use attractor seeds instead
        of attractor candidates to check for motif-avoidant attractors. This means that the attractor
        detection can take much longer. In particular, it will not expand the SD if it cannot detect
        all attractors, meaning it can get "stuck". However, assuming you want to run full attractor
        detection anyway, this might save you some time, as you won't need to re-run  the failed
        candidate detection. (This is only relevant for models where the attractors are very
        complex and the candidate state detection can fail with default settings).
        """
        return expand_source_blocks(
            self,
            find_motif_avoidant_attractors,
            size_limit=size_limit,
            optimize_source_nodes=optimize_source_nodes,
            check_maa_exact=exact_attractor_detection,
        )

    def expand_bfs(
        self,
        node_id: int | None = None,
        bfs_level_limit: int | None = None,
        size_limit: int | None = None,
    ) -> bool:
        """
        Explore the succession diagram in a BFS manner.

        If `node_id` is given, initiate BFS from this node. Otherwise use root.
        If `bfs_level_limit` is given, this is the last "level" (distance from
        the initial node) of nodes that should be expanded (any subsequent child
        nodes are left unexplored).

        If `size_limit` is given, the procedure stops once `SuccessionDiagram`
        exceeds the given size.

        With default settings, the method will explore the whole succession
        diagram without any restrictions.

        The method returns `True` if the whole exploration was completed and
        `False` if it was terminated early based on one of the aforementioned
        conditions.

        Note that the procedure also explores nodes that are already expanded.
        I.e. if all nodes at levels 0,1,2 are expanded, but there are stub nodes
        on level 3, the procedure will still discover and expand these stub
        nodes (assuming sufficient level and size limit).

        Also note that the `size_limit` is only a soft limit: for each node, we
        always have to create all child nodes when expanding it. Hence the
        procedure can only check the condition in between expanding new nodes.
        """
        return expand_bfs(self, node_id, bfs_level_limit, size_limit)

    def expand_dfs(
        self,
        node_id: int | None = None,
        dfs_stack_limit: int | None = None,
        size_limit: int | None = None,
    ) -> bool:
        """
        Similar to `expand_bfs`, but uses DFS instead of BFS.

        The only major difference is the `dfs_stack_limit` which restricts the
        size of the DFS stack. Nodes that would appear "deeper" in the stack
        than this limit are left unexpanded. Note that this stack size is
        technically *some* form of distance from the initial node, but not
        necessarily the minimal distance.
        """
        return expand_dfs(self, node_id, dfs_stack_limit, size_limit)

    def expand_minimal_spaces(
        self,
        node_id: int | None = None,
        size_limit: int | None = None,
        skip_ignored: bool = False,
    ) -> bool:
        """
        Expands the succession diagram in a way that guarantees every minimal
        trap space to be reachable from the root node, but otherwise (greedily)
        avoids unnecesary expansion of nodes whenever possible.

        The algorithm is loosely based on `expand_bfs` implementation, but on
        each BFS level only expands the first node that still contains some
        minimal trap space not covered by a previously expanded node at that
        level.

        The resulting succession diagram construction is deterministic, but can
        vary if some nodes are already expanded initially. In such case, the
        procedure still tries to avoid expanding unnecessary nodes, which means
        existing expanded nodes can be prioritised over the "canonical" ones.

        Optionally, you can start the expansion from a specific node that is not
        the root using `node_id`, or restrict the size of the succession diagram
        with `size_limit`.

        Returns `True` if the expansion procedure terminated without exceeding
        the size limit.

        If `skip_ignored` is set, any nodes that are not expanded by this procedure
        are "skipped" with edges redirected to the corresponding minimal trap spaces
        (see also `SuccessionDiagram.skip_remaining`). This is usually faster than
        using `skip_remaining` directly, since the minimal trap spaces are only
        computed once. However, note that if used with `size_limit`, nodes that
        are not processed when `size_limit` is reached remain unprocessed
        (i.e. it is not guaranteed that all nodes are either fully expanded or skipped).
        """
        return expand_minimal_spaces(self, node_id, size_limit, skip_ignored)

    def expand_attractor_seeds(self, size_limit: int | None = None) -> bool:
        """
        Expands the succession diagram such that for every asynchronous
        attractor, there is at least one expanded trap space which is the
        minimal trap space containing this attractor. In other words, the
        procedure expands the succession diagram as little as possible, but
        ensures that every attractor is "as easy to identify" as possible.

        After this procedure, it is sufficient to search for attractors in
        expanded nodes. Note that this method does not perform exact attractor
        identification. It is possible that some nodes are expanded spuriously
        and the succession diagram is thus larger than necessary.
        """
        return expand_attractor_seeds(self, size_limit)

    def expand_to_target(
        self, target: BooleanSpace, size_limit: int | None = None
    ) -> bool:
        """
        Expands the succession diagram using BFS in such a way that only nodes
        which intersect `target` but are not fully contained in it are expanded.

        This is used for example in control, as it ensures that all possible
        branches of the succession diagram relevant for a particular "target
        subspace" are expanded as much as necessary, but not more.
        """
        return expand_to_target(self, target, size_limit)

    def skip_to_minimal(self, node_id: int) -> bool:
        """
        Skip the expansion of this node (see also `NodeData.skipped`) and add extra
        edges that connect it directly to its minimal trap spaces. Returns `False`
        if the node is already expanded.

        Note that this method is relatively inefficient when applied to multiple
        nodes repeatedly, as it has to recompute the minimal trap spaces for each node.
        To turn multiple nodes into skip nodes, see also `skip_remaining`.
        """

        node = self.node_data(node_id)

        if node["expanded"]:
            return False

        pn = self.node_percolated_petri_net(node_id, compute=True)
        minimal_traps = trappist(network=pn, problem="min")
        minimal_traps = [(node["space"] | x) for x in minimal_traps]

        if len(minimal_traps) == 1 and minimal_traps[0] == node["space"]:
            # This node is a minimal trap space
            # and thus cannot be skipped.
            node["expanded"] = True
            return True

        for m_trap in minimal_traps:
            m_id = self._ensure_node(node_id, m_trap)
            # Also expand the minimal trap space, since we know
            # it has no successors.
            m_data = self.node_data(m_id)
            m_data["expanded"] = True

        node["expanded"] = True
        node["skipped"] = True

        if self.config["debug"]:
            print(f"[{node_id}] Added {len(minimal_traps)} skip edges.")

        return True

    def skip_remaining(self) -> int:
        """
        Apply `skip_to_minimal` to every node that is not expanded.

        This is faster than calling the method individually if the number of
        nodes is high since we can cache the minimal trap spaces.

        Returns the number of created skip nodes.
        """

        pn = self.node_percolated_network(self.root(), compute=True)
        root_space = self.node_data(self.root())["space"]
        minimal_traps = trappist(network=pn, problem="min")
        minimal_traps = [root_space | x for x in minimal_traps]

        if self.config["debug"]:
            print(f"Skipping remaining nodes. Found {len(minimal_traps)} trap spaces.")

        trap_with_id: list[tuple[int, BooleanSpace]] = []
        for m_trap in minimal_traps:
            m_id = self._ensure_node(None, m_trap)
            m_data = self.node_data(m_id)
            trap_with_id.append((m_id, m_trap))
            # We can expand the minimal trap spaces since
            # they don't have successors.
            m_data["expanded"] = True

        skipped_nodes = 0
        for node_id in self.node_ids():
            node = self.node_data(node_id)

            if node["expanded"]:
                continue

            skip_edges = 0
            for m_id, m_trap in trap_with_id:
                if is_subspace(m_trap, node["space"]):
                    self._ensure_edge(node_id, m_id, m_trap)
                    skip_edges += 1

            node["skipped"] = True
            node["expanded"] = True
            skipped_nodes += 1

            # At this point, all minimal traps must be expanded,
            # hence we should never skip one.
            assert not self.node_is_minimal(node_id)

        if self.config["debug"]:
            print(f"Skipped {skipped_nodes} nodes.")

        return skipped_nodes

    def _update_node_depth(self, node_id: int, parent_id: int):
        """
        An internal method that updates the depth of a node based on a specific
        parent node. This assumes that there is an edge from `parent` to
        `node_id`.

        Note that the depth can only increase.
        """
        assert self.dag.edges[parent_id, node_id] is not None
        parent_depth = cast(int, self.dag.nodes[parent_id]["depth"])
        current_depth = cast(int, self.dag.nodes[node_id]["depth"])
        self.dag.nodes[node_id]["depth"] = max(current_depth, parent_depth + 1)

    def _expand_one_node(self, node_id: int):
        """
        An internal method that expands a single node of the succession diagram.

        This entails computing the maximal trap spaces within the node (stable
        motifs) and creating a node for the result (if it does not exist yet).

        If the node is already expanded, the method does nothing.

        If there are already some attractor data for this node (stub nodes can
        have associated attractor data), this data is erased.
        """
        node = cast(dict[str, Any], self.dag.nodes[node_id])
        if node["expanded"]:
            return

        # If the node had any attractor data computed as unexpanded, these are
        # no longer valid and need to be erased.
        node["attractor_seeds"] = None
        node["attractor_candidates"] = None
        node["attractor_sets"] = None

        current_space = node["space"]

        if self.config["debug"]:
            print(
                f"[{node_id}] Expanding: {len(self.node_data(node_id)['space'])} fixed vars."
            )

        if len(current_space) == self.network.variable_count():
            # This node is a fixed-point. Trappist would just
            # return this fixed-point again. No need to continue.
            if self.config["debug"]:
                print(f"[{node_id}] Found fixed-point: {current_space}.")
            node["expanded"] = True
            return

        # We use the non-propagated Petri net for backwards-compatibility reasons here.
        # The SD created from the restricted Petri net is technically correct, but can
        # propagate some of the input values further and yields a smaller SD.
        source_nodes = []
        if node_id == self.root():
            source_nodes = extract_source_variables(self.petri_net)

        sub_spaces: list[BooleanSpace]

        # Only use the percolated PN if it is already known.
        pn = node["percolated_petri_net"]
        if pn is not None:
            # We have a pre-propagated PN for this sub-space, hence we can use
            # that to compute the trap spaces.
            partial_sub_spaces = trappist(
                pn,
                problem="max",
                optimize_source_variables=source_nodes,
                solution_limit=self.config["max_motifs_per_node"],
            )
            sub_spaces = [(s | current_space) for s in partial_sub_spaces]
        else:
            # If we (for whatever reason) don't have the pre-propagated PN,
            # we can still use the "global" PN and let trappist deal with the restriction.
            sub_spaces = trappist(
                self.petri_net,
                problem="max",
                ensure_subspace=current_space,
                optimize_source_variables=source_nodes,
                solution_limit=self.config["max_motifs_per_node"],
            )

        # Release the Petri net once the sub_spaces are computed.
        # It might be needed later for attractor computation, but it
        # uses a lot of memory in large diagrams to keep all the nets
        # in memory.
        node["percolated_petri_net"] = None

        if len(sub_spaces) == self.config["max_motifs_per_node"]:
            raise RuntimeError(
                f"Exceeded the maximum amount of stable motifs per node ({self.config['max_motifs_per_node']}; see `SuccessionDiagramConfiguration.max_motifs_per_node`)."
            )

        # Sort the spaces based on a unique key in case trappist is not always
        # sorted deterministically.
        sub_spaces = sorted(
            sub_spaces, key=lambda space: space_unique_key(space, self.network)
        )

        if len(sub_spaces) == 0:
            if self.config["debug"]:
                print(f"[{node_id}] Found minimum trap space: {current_space}.")
            node["expanded"] = True
            return

        if self.config["debug"]:
            print(f"[{node_id}] Found sub-spaces: {len(sub_spaces)}")

        for sub_space in sub_spaces:
            self._ensure_node(node_id, sub_space)

        # If everything else worked out, we can mark the node as expanded.
        node["expanded"] = True

    def _ensure_node(self, parent_id: int | None, stable_motif: BooleanSpace) -> int:
        """
        Internal method that ensures the provided node is present in this
        succession diagram as a child of the given `parent_id`.

        The `stable_motif` is an "initial" trap space that is then percolated to
        compute the actual fixed variables for this node. The method also
        updates the depth of the child node if necessary.

        If the `parent_id` is not given, no edge is created and depth is
        considered to be zero (i.e. the node is the root).
        """

        fixed_vars = percolate_space(self.symbolic, stable_motif)

        key = space_unique_key(fixed_vars, self.network)

        child_id = None
        if key not in self.node_indices:
            child_id = self.dag.number_of_nodes()

            # Note: this must match the fields of the `NodeData` class
            self.dag.add_node(  # type: ignore
                child_id,
                space=fixed_vars,
                depth=0,
                expanded=False,
                percolated_network=None,
                percolated_petri_net=None,
                percolated_nfvs=None,
                attractor_candidates=None,
                attractor_seeds=None,
                attractor_sets=None,
                parent_node=parent_id,
                skipped=None,
            )
            self.node_indices[key] = child_id
        else:
            child_id = self.node_indices[key]

        assert child_id is not None

        if parent_id is not None:
            self._ensure_edge(parent_id, child_id, stable_motif)

        return child_id

    def _ensure_edge(self, parent_id: int, child_id: int, stable_motif: BooleanSpace):
        # TODO: It seems that there are some networks where the same child
        # can be reached through multiple stable motifs. Not sure how to
        # approach these... but this is probably good enough for now.
        if not self.dag.has_edge(parent_id, child_id):  # type: ignore
            self.dag.add_edge(parent_id, child_id, motif=stable_motif, all_motifs=[stable_motif])  # type: ignore
        else:
            self.dag.edges[parent_id, child_id]["all_motifs"].append(stable_motif)  # type: ignore
        self._update_node_depth(child_id, parent_id)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from biobalm.succession_diagram import SuccessionDiagram

from biobalm.space_utils import intersect, is_subspace
from biobalm.types import BooleanSpace


def expand_to_target(
    sd: SuccessionDiagram, target: BooleanSpace, size_limit: int | None = None
):
    """
    See `SuccessionDiagram.exapnd_to_target` for documentation.
    """

    root = sd.root()
    seen = set([root])

    level_id = 0
    current_level = [root]
    next_level: list[int] = []

    while len(current_level) > 0:
        for node in current_level:
            node_space = sd.node_data(node)["space"]

            if intersect(node_space, target) is None:
                # If `node_space` does not intersect with `target`, it is not relevant
                # and we can safely keep it as unexpanded "stub".
                continue

            if is_subspace(node_space, target) and not node_space == target:
                # If `node_space` is a subspace of `target`, it is relevant but expanding
                # it will not add any new information, we can thus also keep it unexpanded.
                continue

            # Check if the size limit has been exceeded already.
            if (size_limit is not None) and (len(sd) >= size_limit):
                # Size limit reached.
                return False

            # Compute successors if necessary.
            successors = sd.node_successors(node, compute=True)
            # Sort successors to avoid non-determinism.
            successors = sorted(successors)

            # Add successors to the next level and to the seen set.
            for s in successors:
                if s not in seen:
                    seen.add(s)
                    next_level.append(s)

        # If not, "move on" to the next level.
        level_id += 1
        current_level = next_level
        next_level = []

    return True



"""
This module implementes methods for permanent target control of Boolean networks
based on the structure of a succession diagram.
"""

from __future__ import annotations

from functools import reduce
from itertools import combinations, product
from typing import Iterator, Literal, cast

import networkx as nx  # type: ignore
from biodivine_aeon import AsynchronousGraph, BooleanNetwork

from biobalm.space_utils import intersect, is_subspace, percolate_space
from biobalm.succession_diagram import SuccessionDiagram
from biobalm.types import BooleanSpace, ControlOverrides, SubspaceSuccession


class Intervention:
    def __init__(
        self,
        control: list[ControlOverrides],
        strategy: Literal["internal", "all"],
        succession: SubspaceSuccession,
    ):
        """A class for encoding an intervention to a network to reach a target subspace.

        Generally, this class is created by the
        :func:`succession_control<biobalm.control.succession_control>` function,
        which returns a list of `Intervention` objects. Manipulating the
        contents of these objects is only recommended for advanced use cases.
        Typically, it is sufficient to print this object to see a human-readable
        explanation of how to interpret the intervention.

        Note that two interventions are considered to be equal if they act on the same succession with
        equal controls applied to each subspace in the succession. Thus, two
        interventions that override the same nodes in the same way can be
        unequal, even when applied to the same subspaces. This is because each
        subspace imposes a duration condition on the overrides, i.e., the
        override must be maintained until the succession subspace is reached.
        Changing the order of the subspaces considered can alter this stop
        condition.

        Parameters
        ----------
        control : list[ControlOverrides]
            The :class:`ControlOverrides<biobalm.types.ControlOverrides>` objects,
            in order, that are applied. The order of the list corresponds the
            order of the subspaces in the succession. Each
            :class:`ControlOverrides<biobalm.types.ControlOverrides>` object
            represens a list of overrides, stored as a dictionary of node-value
            pairs, that drive the system to the corresponding subspace. Each
            :class:`ControlOverrides<biobalm.types.ControlOverrides>` object is
            sorted by key value (i.e., alphabetically) upon creation of the
            `Intervention` object to maintain a canonical ordering.
        strategy : str
            Either "internal" or "all"; "internal" means that the
            :class:`ControlOverrides<biobalm.types.ControlOverrides>`
        succession : SubspaceSuccession
            A sequence of subspaces that are targeted by the corresponding
            entries of `control`.

        Example
        -------
        >>> import biobalm
        >>> sd = biobalm.SuccessionDiagram.from_rules(
        ...     \"\"\"
        ...     A, B & C
        ...     B, A & C
        ...     C, A & B
        ...     \"\"\"
        ...     )
        >>> target = {"A": 1, "B": 1, "C": 1}
        >>> interventions = biobalm.control.succession_control(sd, target)
        >>> intervention = interventions[0] # only one in this case
        >>> intervention.control
        [[{'A': 1, 'B': 1}, {'A': 1, 'C': 1}, {'B': 1, 'C': 1}]]
        >>> intervention.strategy
        'internal'
        >>> intervention.succession
        [{'A': 1, 'B': 1, 'C': 1}]
        >>> intervention.successful
        True
        >>> print(intervention)
        Intervention is SUCCESSFUL operating on
        {'A': 1, 'B': 1, 'C': 1}
        override
        ({'A': 1, 'B': 1} or {'A': 1, 'C': 1} or {'B': 1, 'C': 1})
        """

        # we store the control in a canonical (sorted) representation
        self._control: list[ControlOverrides] = []
        for c in control:
            cs = sorted(map(lambda x: sorted(x.items()), c))
            self._control.append(list(map(dict, cs)))  # type: ignore

        self._strategy = strategy
        self._succession = succession
        self._successful = not any(not c for c in control)

    @property
    def control(self):
        return self._control

    @property
    def strategy(self):
        return self._strategy

    @property
    def succession(self):
        return self._succession

    @property
    def successful(self):
        return self._successful

    def __eq__(self, other: object):
        if not isinstance(other, Intervention):
            return False

        if self.succession != other.succession:
            return False

        if len(self.control) != len(other.control):
            return False

        for d1, d2 in zip(self.control, other.control):
            if not controls_are_equal(d1, d2):
                return False

        return True

    def __repr__(self):
        return (
            f"Intervention("
            f"{self.control},"
            f"{self.strategy},"
            f"{self.succession},"
            f"{self.successful})"
        )

    def __str__(self):
        succession_string = (
            f"Intervention is {'' if self.successful else 'UN'}SUCCESSFUL operating on\n"
            + "\n".join(map(str, self.succession))
            + "\noverride\n"
        )
        if self.strategy == "internal":
            return succession_string + " and then\n".join(
                f"({' or '.join(map(str, motif_control))})"
                for motif_control in self.control
            )
        elif self.strategy == "all":
            return succession_string + " temporarily, and then\n".join(
                f"({' or '.join(map(str, motif_control))})"
                for motif_control in self.control
            )
        else:
            return "unknown strategy: " + self.__repr__()

    def all_control_strategies(self) -> Iterator[ControlOverrides]:
        """
        Returns all possible combinations of `ControlOverrides` sequences that
        can be used to execute this `Intervention`.

        Internally, an intervention consists of multiple control steps that
        need to be taken sequentially. For each step in the sequence, an intervention
        can have multiple options of how to execute it. With this method,
        we can generate the actual sequences that arise by combining all the
        available options for each step.
        """
        return map(lambda x: list(x), product(*self._control))


def succession_control(
    succession_diagram: SuccessionDiagram,
    target: BooleanSpace,
    strategy: Literal["internal", "all"] = "internal",
    max_drivers_per_succession_node: int | None = None,
    forbidden_drivers: set[str] | None = None,
    successful_only: bool = True,
    skip_feedforward_successions: bool = False,
) -> list[Intervention]:
    """
    Performs succession-diagram control to reach a target subspace.

    Parameters
    ----------
    succession_diagram : SuccessionDiagram
        The succession diagram from which successions and rules will be
        extracted.
    target : BooleanSpace
        The target subspace.
    strategy : str, optional
        The searching strategy to use to look for driver nodes. Options are
        'internal' (default) and 'all'.
    max_drivers_per_succession_node: int | None = None,
        The maximum number of drivers that will be tested for a succession
        diagram node. If `None`, then a number of drivers up to the size of the
        succession diagram node's stable motif will be tested
    forbidden_drivers: set[str] | None
        A set of forbidden drivers that will not be overridden for control. If
        `None`, then all nodes are candidates for control.
    successful_only: bool
        Whether to only return successful interventions (default: `True`).

    Returns
    -------
    list[Intervention]
        A list of control :class:`Intervention<biobalm.control.Intervention>`
        objects. Note that when `successful_only` is `False`, returned
        interventions may be unsuccessful if `max_drivers_per_succession_node`
        is set too small, or crucial nodes are included in `forbidden_drivers`.
        To test, examine the `successful` property of the intervention.

    Example
    -------
    >>> import biobalm
    >>> from biobalm.control import succession_control
    >>> sd = biobalm.SuccessionDiagram.from_rules(
    ...     \"\"\"
    ...     S, S
    ...     A, S | B
    ...     B, A
    ...     C, A | D
    ...     D, C
    ...     E, false
    ...     \"\"\"
    ...     )
    >>> target = {"S": 0, "E": 0, "A": 0, "B": 0, "C": 1, "D": 1}
    >>> interventions = succession_control(sd, target, forbidden_drivers = {"A"})
    >>> interventions.sort(key=lambda x: len(x.control)) # to maintain fixed order
    >>> for i in interventions:
    ...     print(f'{i}\\n'+'-'*20)
    ...
    Intervention is SUCCESSFUL operating on
    {'S': 0}
    {'A': 0, 'B': 0}
    {'C': 1, 'D': 1}
    override
    ({'S': 0}) and then
    ({'B': 0}) and then
    ({'C': 1} or {'D': 1})
    --------------------
    Intervention is SUCCESSFUL operating on
    {'S': 0}
    {'C': 1, 'D': 1}
    {'A': 0, 'B': 0}
    override
    ({'S': 0}) and then
    ({'C': 1} or {'D': 1}) and then
    ({'B': 0})
    --------------------

    Example
    -------
    >>> import biobalm
    >>> from biobalm.control import succession_control
    >>> sd = biobalm.SuccessionDiagram.from_rules(
    ...         \"\"\"
    ...     S, S
    ...     A, S | B
    ...     B, A
    ...     C, A | D
    ...     D, C
    ...     E, false
    ...     \"\"\"
    ...     )
    >>> target = {"S": 0, "E": 0, "A": 0, "B": 0, "C": 1, "D": 1}
    >>> interventions = succession_control(sd, target, strategy = "all")
    >>> interventions.sort(key=lambda x: len(x.control)) # to maintain fixed order
    >>> for i in interventions:
    ...     print(f'{i}\\n'+'-'*20)
    ...
    Intervention is SUCCESSFUL operating on
    {'S': 0}
    {'A': 0, 'B': 0}
    {'C': 1, 'D': 1}
    override
    ({'S': 0}) temporarily, and then
    ({'A': 0} or {'B': 0}) temporarily, and then
    ({'C': 1} or {'D': 1})
    --------------------
    Intervention is SUCCESSFUL operating on
    {'S': 0}
    {'C': 1, 'D': 1}
    {'A': 0, 'B': 0}
    override
    ({'S': 0}) temporarily, and then
    ({'A': 1} or {'B': 1} or {'C': 1} or {'D': 1}) temporarily, and then
    ({'A': 0} or {'B': 0})
    --------------------
    """
    interventions: list[Intervention] = []

    successions = successions_to_target(
        succession_diagram,
        target=target,
        expand_diagram=True,
        skip_feedforward_successions=skip_feedforward_successions,
    )

    for succession in successions:
        controls = drivers_of_succession(
            succession_diagram.symbolic,
            succession,
            strategy=strategy,
            max_drivers_per_succession_node=max_drivers_per_succession_node,
            forbidden_drivers=forbidden_drivers,
        )
        intervention = Intervention(controls, strategy, succession)

        if not successful_only or intervention.successful:
            interventions.append(intervention)

    return interventions


def successions_to_target(
    succession_diagram: SuccessionDiagram,
    target: BooleanSpace,
    expand_diagram: bool = True,
    skip_feedforward_successions: bool = False,
) -> list[SubspaceSuccession]:
    """Find lists of nested trap spaces (successions) that lead to the
    specified target subspace.

    Generally, it is not necessary to call this function directly, as it is
    automatically invoked by
    :func:`succession_control<biobalm.control.succession_control>`. It is
    primarily provided in the public API for testing and benchmarking purposes,
    or in the case that the user wants to implement a custom strategy to
    identify succession drivers rather than relying on
    :func:`drivers_of_succession<biobalm.control.drivers_of_succession>`.

    Parameters
    ----------
    succession_diagram : SuccessionDiagram
        The succession diagram from which successions will be extracted.
    target : BooleanSpace
        The target subspace.
    expand_diagram: bool
        Whether to ensure that the succession diagram is expanded enough to
        capture all paths to the target (default: True).
    skip_feedforward_successions: bool
        Whether to skip redundant successions (default: False). Skipping these
        can reduce the number of interventions to test, yielding performance
        improvements, but can also cause the algorithm to miss some valid
        interventions, particularly in cases when the order of intervention
        application is important.

    Returns
    -------
    list[SubspaceSuccession]
        A list of successions, where each succession is a list of sequentially
        nested trap spaces that specify the target.
    """
    successions: list[SubspaceSuccession] = []
    # Tracks the combined perturbation that needs to be applied for the whole
    # succession to take effect. We use this to detect which successions are
    # redundant when they can be replaced by a succession with a subset
    # signature. Used when skip_feedforward_successions is True
    succession_signatures: list[BooleanSpace] = []

    # expand the succession_diagram toward the target
    if expand_diagram:
        succession_diagram.expand_to_target(
            target=target,
        )

    # these contradict the target or have a motif avoidant attractor that isn't
    # in full agreement with the target
    hot_lava_nodes: set[int] = set()

    descendant_map: dict[int, set[int]] = {}
    for s in succession_diagram.node_ids():
        is_consistent = intersect(succession_diagram.node_data(s)["space"], target)
        is_goal = is_subspace(succession_diagram.node_data(s)["space"], target)
        is_minimal = succession_diagram.node_is_minimal(s)
        if not is_consistent or (not is_goal and is_minimal):
            hot_lava_nodes.add(s)

        descendant_map[s] = set(nx.descendants(succession_diagram.dag, s))  # type: ignore
        descendant_map[s].add(s)  # for our purposes, s is its own descendant
    found_valid_target_node = False
    for s in succession_diagram.node_ids():
        # a node is a valid end point if all
        # 1) its descendents (including itself) are not "hot lava"
        # 2) it has a parent that is or can reach "hot lava" (otherwise, we can
        # just control to the parent) note that all nodes are either cold lava
        # or hot lava, but not both or neither
        if descendant_map[s] & hot_lava_nodes:
            continue
        found_valid_target_node = True
        if not any(
            descendant_map[p] & hot_lava_nodes
            for p in succession_diagram.dag.predecessors(s)  # type: ignore
        ):
            continue

        for path in cast(
            list[list[int]],
            nx.all_simple_paths(  # type: ignore
                succession_diagram.dag,
                source=succession_diagram.root(),
                target=s,
            ),
        ):
            motif_list = [
                succession_diagram.edge_all_stable_motifs(x, y, reduced=True)
                for x, y in zip(path[:-1], path[1:])
            ]
            for succession_tuple in product(*motif_list):
                succession = list(succession_tuple)
                if skip_feedforward_successions:
                    signature = reduce(lambda x, y: x | y, succession)
                    # First, check if any existing successions can be eliminated
                    # because they are redundant w.r.t. to this succession.
                    # (`reversed` is important here, because that way a delete
                    # only impacts indices that we already processed)
                    skip_completely = False
                    for i in reversed(range(len(succession_signatures))):
                        existing_signature = succession_signatures[i]
                        if is_subspace(signature, existing_signature):
                            # The current `path` is already superseded by a path in successions.
                            skip_completely = True
                            break
                        if is_subspace(existing_signature, signature):
                            # A path in successions is made redundant by the current path.
                            del succession_signatures[i]
                            del successions[i]
                    if skip_completely:
                        continue

                    succession_signatures.append(signature)
                successions.append(succession)

    if found_valid_target_node and len(successions) == 0:
        successions = [[]]

    return successions


def drivers_of_succession(
    bn: BooleanNetwork | AsynchronousGraph,
    succession: list[BooleanSpace],
    strategy: str = "internal",
    max_drivers_per_succession_node: int | None = None,
    forbidden_drivers: set[str] | None = None,
) -> list[ControlOverrides]:
    """
    Find driver nodes of a list of sequentially nested trap spaces.

    Generally, it is not necessary to call this function directly, as it is
    automatically invoked by
    :func:`succession_control<biobalm.control.succession_control>`. It is primarily
    provided in the public API for testing and benchmarking purposes.

    Parameters
    ----------
    bn : BooleanNetwork | AsynchronousGraph
        The network to analyze, which contains the Boolean update functions. Ideally,
        the network should be already provided as a symbolic `AsynchronousGraph`.
        Alternatively, a "raw" `BooleanNetwork` can be provided and the symbolic
        `AsynchronousGraph` is created automatically.
    succession : list[BooleanSpace]
        A list of sequentially nested trap spaces that specify the target.
    strategy: str
        The searching strategy to use to look for driver nodes. Options are
        'internal' (default) and 'all'.
    max_drivers_per_succession_node: int | None = None,
        The maximum number of drivers that will be tested for a succession
        diagram node. If `None`, then a number of drivers up to the size of the
        succession diagram node's stable motif will be tested.
    forbidden_drivers: set[str] | None
        A set of forbidden drivers that will not be overridden for control. If
        `None`, then all nodes are candidates for control.

    Returns
    -------
    list[ControlOverrides]
        A list of controls. Each control is a list of lists of driver sets,
        represented as state dictionaries. Each list item corresponds to a list
        of drivers for the corresponding trap space in the succession.
    """
    if isinstance(bn, BooleanNetwork):
        bn = AsynchronousGraph(bn)

    control_strategies: list[ControlOverrides] = []
    assume_fixed: BooleanSpace = {}
    for ts in succession:
        control_strategies.append(
            find_drivers(
                bn,
                ts,
                strategy=strategy,
                assume_fixed=assume_fixed,
                max_drivers_per_succession_node=max_drivers_per_succession_node,
                forbidden_drivers=forbidden_drivers,
            )
        )
        ldoi = percolate_space(bn, ts | assume_fixed)
        assume_fixed.update(ldoi)

    return control_strategies


def find_drivers(
    bn: BooleanNetwork | AsynchronousGraph,
    target_trap_space: BooleanSpace,
    strategy: str = "internal",
    assume_fixed: BooleanSpace | None = None,
    max_drivers_per_succession_node: int | None = None,
    forbidden_drivers: set[str] | None = None,
) -> ControlOverrides:
    """
    Finds drives of a given target trap space

    Generally, it is not necessary to call this function directly, as it is
    automatically invoked by
    :func:`drivers_of_succession<biobalm.control.drivers_of_succession>`, which in
    turn is invoked by
    :func:`succession_control<biobalm.control.succession_control>`. It is primarily
    provided in the public API for testing and benchmarking purposes.

    Parameters
    ----------
    bn : BooleanNetwork | AsynchronousGraph
        The network to analyze, which contains the Boolean update functions. Ideally,
        the network should be already provided as a symbolic `AsynchronousGraph`.
        Alternatively, a "raw" `BooleanNetwork` can be provided and the symbolic
        `AsynchronousGraph` is created automatically.
    target_trap_space : BooleanSpace
        The trap space we want to find drivers for.
    strategy: str
        The searching strategy to use to look for driver nodes. Options are
        'internal' (default) and 'all'.
    assume_fixed: dict[str,int] | None
        A dictionary of fixed variables that should be assumed to be fixed.
    max_drivers_per_succession_node: int | None = None,
        The maximum number of drivers that will be tested for a succession
        diagram node. If `None`, then a number of drivers up to the size of the
        succession diagram node's stable motif will be tested.
    forbidden_drivers: set[str] | None
        A set of forbidden drivers that will not be overridden for control. If
        `None`, then all nodes are candidates for control.

    Returns
    -------
    ControlOverrides
        A list of internal driver sets, represented as state dictionaries. If
        empty, then no drivers are found. This can happen if
        `max_drivers_per_succession_node` is not `None`, or if all controls
        require nodes in `forbidden_drivers`.
    """
    if isinstance(bn, BooleanNetwork):
        bn = AsynchronousGraph(bn)

    if assume_fixed is None:
        assume_fixed = {}
    if forbidden_drivers is None:
        forbidden_drivers = set()

    target_trap_space_inner = {
        k: v for k, v in target_trap_space.items() if k not in assume_fixed
    }

    if strategy == "internal":
        driver_pool = set(target_trap_space_inner) - forbidden_drivers
    elif strategy == "all":
        driver_pool = set(bn.network_variable_names()) - forbidden_drivers
    else:
        raise ValueError("Unknown driver search strategy")

    if max_drivers_per_succession_node is None:
        max_drivers_per_succession_node = len(target_trap_space_inner)

    drivers: ControlOverrides = []
    for driver_set_size in range(max_drivers_per_succession_node + 1):
        for driver_set in combinations(driver_pool, driver_set_size):
            if any(set(d) <= set(driver_set) for d in drivers):
                continue

            if strategy == "internal":
                driver_dict: BooleanSpace = {
                    k: cast(Literal[0, 1], target_trap_space_inner[k])
                    for k in driver_set
                }
                ldoi = percolate_space(bn, driver_dict | assume_fixed)
                if target_trap_space.items() <= ldoi.items():
                    drivers.append(driver_dict)
            elif strategy == "all":
                for vals in product([0, 1], repeat=driver_set_size):
                    driver_dict = {
                        driver: cast(Literal[0, 1], value)
                        for driver, value in zip(driver_set, vals)
                    }
                    ldoi = percolate_space(bn, driver_dict | assume_fixed)
                    if target_trap_space.items() <= ldoi.items():
                        drivers.append(driver_dict)
    return drivers


def controls_are_equal(a: ControlOverrides, b: ControlOverrides) -> bool:
    """
    Determine if two :class:`ControlOverrides<biobalm.types.ControlOverrides>`
    objects are equal.

    Two `ControlOverrides` objects are equal if they contain the same
    :class:`BooleanSpace<biobalm.types.BooleanSpace>` objects, regardless of their
    ordering.

    Parameters
    ----------
    a : ControlOverrides
        First :class:`ControlOverrides<biobalm.types.ControlOverrides>` object for
        comparison.
    b : ControlOverrides
        Second :class:`ControlOverrides<biobalm.types.ControlOverrides>` object for
        comparison.

    Returns
    -------
    bool
        Returns `True` if the two
        :class:`ControlOverrides<biobalm.types.ControlOverrides>` objects are
        equal.
    """
    return set(frozenset(x.items()) for x in a) == set(frozenset(x.items()) for x in b)



from __future__ import annotations

import itertools as it
from typing import TYPE_CHECKING, Callable, cast

from biobalm.space_utils import percolate_network
from biobalm.types import BooleanSpace
from biobalm.interaction_graph_utils import source_nodes

if TYPE_CHECKING:
    from biobalm.succession_diagram import SuccessionDiagram

    ExpanderFunctionType = Callable[[SuccessionDiagram], bool]


def expand_source_SCCs(
    sd: SuccessionDiagram,
    check_maa: bool,
    recursion: int = 0,
    expander: ExpanderFunctionType | None = None,
) -> bool:
    """
    1. percolate
    2. find source nodes, fix combinations and then percolate
    3. at every level, find all source SCCs, expand them
    4. when there are no more SCCs, expand in a usual way.

    TODO: it will be rare especially for empirical models,
    but there can be source SCCs even before fixing source node combinations.
    It could be useful to find them first, calculate scc_sd and store them somewhere,
    rather than calculating the same thing for every source node combination

    Parameters
    ----------
    check_maa - if True, check for motif avoidant attractors when
                expanding the sd with source SCCs.
                This grants significant speedup when searching for all attractors.
                if False, ignores motif avoidant attractors in source SCCs.

    """

    if expander is None:
        # By default, we recursively call the SCC expansion.
        def default_expander(sd: SuccessionDiagram):
            return expand_source_SCCs(sd, check_maa, recursion + 1)

        expander = default_expander

    if sd.config["debug"]:
        print(
            f"Start SD expansion using SCC decomposition on {sd.network}, recursion level {recursion}."
        )

    root = sd.root()

    # Usage of sets prevents node repetition in levels (this can happen if we independently
    # percolate to the same "downstream" node after fixing different source SCC values).
    current_level: set[int] = set([root])
    next_level: set[int] = set()

    # This already accounts for constant percolation.
    node_space = sd.node_data(root)["space"]

    # find source nodes
    perc_bn = percolate_network(sd.network, node_space)
    sources = source_nodes(perc_bn)

    if sd.config["debug"]:
        print(f" > Computed source/input variable(s): {sources}")

    # get source nodes combinations and expand root node
    if len(sources) != 0:
        # If there are too many source nodes, this can generate an absurdly large SD.
        # This would be a problem even without the SCC expansion, but we can just
        # stop the whole thing faster because we know how many nodes it generates now.
        if 2 ** len(sources) > sd.config["max_motifs_per_node"]:
            raise RuntimeError(
                f"Exceeded the maximum amount of stable motifs per node ({sd.config['max_motifs_per_node']}; see `SuccessionDiagramConfiguration.max_motifs_per_node`)."
            )
        else:
            if sd.config["debug"]:
                print(
                    f" > Expanding {len(sources)} source node into {2 ** len(sources)} SD nodes."
                )

        bin_values_iter = it.product(range(2), repeat=len(sources))
        for bin_values in bin_values_iter:
            valuation = cast(BooleanSpace, dict(zip(sources, bin_values)))
            sub_space = node_space | valuation

            next_level.add(sd._ensure_node(root, sub_space))  # type: ignore

        # This makes the root artificially "expanded". Also, there
        # can be no attractors here because we are just fixing the source nodes.
        sd.node_data(root)["expanded"] = True
        sd.node_data(root)["attractor_seeds"] = []
        sd.node_data(root)["attractor_sets"] = []
        current_level = next_level
        next_level = set()

    while len(current_level) > 0:
        if sd.config["debug"]:
            print(
                f" > Start SCC expansion of a BFS level with {len(current_level)} node(s)."
            )

        # For each node in the current level, we expand all source SCCs and put the
        # results into a new level.
        for node_id in sorted(current_level):
            source_scc_diagrams = list(sd.source_scc_subdiagrams(node_id))
            if sd.config["debug"]:
                print(
                    f" > [{node_id}] Found {len(source_scc_diagrams)} sub-diagrams while expanding node."
                )

            # If there are no source SCCs, this node is a fixed-point and we can save it
            # into the last level (every other node has at least one source SCC).
            if len(source_scc_diagrams) == 0:
                if sd.config["debug"]:
                    print(f"[{node_id}] > No source SCCs found. Node is a fixed-point.")
                assert len(sd.node_successors(node_id, compute=True)) == 0
                continue

            # If there is only one source SCC, we can do a normal expansion to get to the next level,
            # because the attach sub-diagram process would just expand it a copy it into node_id.
            # Furthermore, if we want to recursively use SCC expansion in SCC expansion, we *have to*
            # do this, otherwise we would just recursively call ourselves forver on a minimal trap spaces,
            # since the recursion wouldn't know where to stop.
            if len(source_scc_diagrams) == 1:
                if sd.config["debug"]:
                    print(f"[{node_id}] > Singe source SCCs found. Expanding normally.")
                next_level = next_level | set(sd.node_successors(node_id, compute=True))
                continue

            attach_at_list: list[int] = [node_id]
            for scc_diagram in source_scc_diagrams:
                fully_expanded = expander(scc_diagram)
                if not fully_expanded:
                    # Something bad happened in the expander function and we can't continue.
                    return False

                if sd.config["debug"]:
                    print(
                        f"[{node_id}] > Source SCC diagram expanded to {len(scc_diagram)} nodes."
                    )

                # At this point, diagram is fully expanded and we can attach its
                # nodes as the successors of `node_id`.
                next_attach_at_list: list[int] = []
                for attach_at in attach_at_list:
                    next_attach_at_list += attach_scc_subdiagram(
                        sd, scc_diagram, attach_at, check_maa
                    )
                attach_at_list = next_attach_at_list

            if attach_at_list == [node_id]:
                # Nothing was attached at this level. This means that all source SCCs
                # have a trivial succession diagram where we can't further "fix" anything else.
                # Usually, this means that we are in a minimal trap space. However, there can also
                # be another SCC connected to the source SCCs that can still have some
                # non-trivial trap space structure.
                #
                # Ideally, we should be able to "ignore" the current source SCCs and move on to the
                # following level of SCCs (because those source SCC stay "unstable" forever), but
                # that is currently not possible, so we just expand normally.
                if sd.config["debug"]:
                    print(
                        f"[{node_id}] > No further nodes attached. Expanding normally."
                    )
                next_level = next_level | set(sd.node_successors(node_id, compute=True))
            else:
                # Otherwise, move everything into the next layer.
                next_level = next_level | set(attach_at_list)

        current_level = next_level
        next_level = set()

    if sd.config["debug"]:
        print(
            f" > SCC expansion terminated with {len(sd)} node(s) on recursion level {recursion}."
        )

    return True


def attach_scc_subdiagram(
    sd: SuccessionDiagram, scc_sd: SuccessionDiagram, attach_at: int, check_maa: bool
) -> list[int]:
    """
    Attach the `scc_sd` to the `attach_at` node of `sd`.

    This means that every node of `scc_sd` will be extended to the full context of `sd`
    and attached as if `attach_at` was the root of `scc_sd`.

    If `check_maa` is set to `True`, the method will also check `scc_sd` for motif-avoidant
    attractors and if MAAs are disproven, the corresponding nodes of `sd` will be also marked
    as not having any MAAs.

    Parameters
    ----------
    sd - The main succession diagram into which we are attaching the `scc_sd`.
    scc_sd - A component subdiagram which will be copied into `sd`.
    attach_at - A node id from `sd` that will be used as the virtual "root" of `scc_sd` while attaching.
    check_maa - A flag that indicates that MAAs absence information should be propagated from the `scc_sd`
                to the main `sd`.

    Returns
    -------
    The IDs of `sd` nodes that correspond to the minimal trap spaces of `scc_sd`.
    """

    if len(scc_sd) == 1:
        # This `scc_sd` has a single minimal SCC, which thus corresponds to the `attach_at` point.
        return [attach_at]

    # Maps node IDs from the `scc_sd` to the extended and copied nodes in `sd`.
    node_id_map: dict[int, int] = {scc_sd.root(): attach_at}

    attach_at_space = sd.node_data(attach_at)["space"]

    # First, copy every node from `scc_sd`` into main `sd`.
    min_traps: list[int] = []
    for scc_node_id in scc_sd.node_ids():
        if scc_node_id == scc_sd.root():
            continue  # Root is implicitly copied.

        scc_node_space = scc_sd.node_data(scc_node_id)["space"]
        extended_node_space = scc_node_space | attach_at_space

        main_node_id = sd._ensure_node(parent_id=None, stable_motif=extended_node_space)  # type: ignore
        node_id_map[scc_node_id] = main_node_id

        if scc_sd.node_is_minimal(scc_node_id):
            min_traps.append(main_node_id)
        else:
            # This node can be marked as expanded, because we know its successors.
            # We just need to add them in the for loop below.
            sd.node_data(main_node_id)["expanded"] = True

        if check_maa:
            if len(scc_sd.node_attractor_candidates(scc_node_id, compute=True)) == 0:
                sd.node_data(main_node_id)["attractor_seeds"] = []
                sd.node_data(main_node_id)["attractor_sets"] = []

    assert len(node_id_map) == len(scc_sd)

    # Then copy all the edges.
    for scc_node_id in scc_sd.node_ids():
        for scc_node_succ in scc_sd.node_successors(scc_node_id):
            inner_stable_motif = scc_sd.edge_stable_motif(scc_node_id, scc_node_succ)

            main_node_id = node_id_map[scc_node_id]
            main_succ_id = node_id_map[scc_node_succ]

            # This should not happen, because the source SCCs are independent.
            # But I don't have a formal proof.
            assert main_node_id != main_succ_id

            sd._ensure_edge(main_node_id, main_succ_id, inner_stable_motif)  # type: ignore

    # This makes the `attach_at` node expanded. We will not be adding new nodes to it later.
    sd.node_data(attach_at)["expanded"] = True
    # Finally, if we are checking for MAAs, we can do that for the root too:
    if check_maa:
        if len(scc_sd.node_attractor_candidates(scc_sd.root(), compute=True)) == 0:
            sd.node_data(attach_at)["attractor_seeds"] = []
            sd.node_data(attach_at)["attractor_sets"] = []

    return min_traps



"""
Utility operations for creating and manipulating symbolic functions.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, cast

if TYPE_CHECKING:
    from biodivine_aeon import Bdd, SymbolicContext

from biodivine_aeon import BddVariableSet, BddValuation

from biobalm.types import BooleanSpace


def state_to_bdd(
    bdd_context: SymbolicContext | BddVariableSet, state: BooleanSpace
) -> Bdd:
    """
    Convert a `BooleanSpace` to a BDD encoding the state singleton.

    Parameters
    ----------
    bdd_context : SymbolicContext | BddVariableSet
        The BDD context in which the state is represented. This is a
        a `biodivine_aeon.BddVariableSet` or `biodivine_aeon.SymbolicContext`
        (which is automatically converted to `biodivine_aeon.BddVariableSet`).
        The context object ensures compatibility between BDDs by
        maintaining a shared collection of variable names and their ordering.
    state : BooleanSpace
        The state to convert to a BDD.

    Returns
    -------
    Bdd
        The BDD encoding the given state.
    """
    bdd = (
        bdd_context
        if isinstance(bdd_context, BddVariableSet)
        else bdd_context.bdd_variable_set()
    )
    return bdd.mk_conjunctive_clause(state)


def state_list_to_bdd(
    bdd_context: SymbolicContext | BddVariableSet, states: list[BooleanSpace]
) -> Bdd:
    """
    Convert a list of state dictionaries to a BDD representation.

    The BDD will be satisfied in any state in the list and in no others.

    Parameters
    ----------
    bdd_context : SymbolicContext | BddVariableSet
        The BDD context in which the state is represented. This is a
        a `biodivine_aeon.BddVariableSet` or `biodivine_aeon.SymbolicContext`
        (which is automatically converted to `biodivine_aeon.BddVariableSet`).
        The context object ensures compatibility between BDDs by
        maintaining a shared collection of variable names and their ordering.
    states : list[BooleanSpace]
        The list of states to convert to a BDD.

    Returns
    -------
    Bdd
        The BDD encoding the given list of states.
    """
    bdd = (
        bdd_context
        if isinstance(bdd_context, BddVariableSet)
        else bdd_context.bdd_variable_set()
    )
    return bdd.mk_dnf(states)


def function_eval(f: Bdd, state: BooleanSpace) -> Literal[0, 1] | None:
    """
    Evaluate a BDD Boolean function in the given state.

    If the state is incomplete (i.e. it is a space), the function value can be
    undetermined. In such case, `None` is returned.

    Parameters
    ----------
    f : Bdd
        The function to evaluate.
    state : BooleanSpace
        The state in which the function is evaluated.

    Returns
    -------
    Literal[0, 1] | None
        The function value in the given state, or `None` if the function is
        undetermined.
    """
    if f.is_false():
        return 0
    if f.is_true():
        return 1

    reduced_f = f.r_restrict(state)
    if reduced_f.is_true():
        return 1
    if reduced_f.is_false():
        return 0
    return None


def function_is_true(f: Bdd, state: BooleanSpace) -> bool:
    """
    `True` if the BDD function evaluates to `1` for the given state (or space).

    Parameters
    ----------
    f : Bdd
        The function to evaluate.
    state : BooleanSpace
        The state in which the function is evaluated.

    Returns
    -------
    bool
        `True` if the function evaluates to `1` in the given state.
    """
    return function_eval(f, state) == 1


def valuation_to_state(ctx: SymbolicContext, valuation: BddValuation) -> BooleanSpace:
    """
    Extract network state from a `BddValuation` into a `BooleanSpace`.

    Parameters
    ----------
    f : SymbolicContext
        A context which maps between network variables and their symbolic counterparts.
    state : BddValuation
        A valuation of the network's symbolic encoding.

    Returns
    -------
    BooleanSpace
        A `BooleanSpace` encoding the state data from the given valuation.
    """

    result: BooleanSpace = {}
    for var, val in valuation.items():
        n_var = ctx.find_network_variable(var)
        if n_var is not None:
            result[ctx.get_network_variable_name(n_var)] = cast(Literal[0, 1], int(val))

    return result



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from biobalm.succession_diagram import SuccessionDiagram


def expand_dfs(
    sd: SuccessionDiagram,
    node_id: int | None = None,
    dfs_stack_limit: int | None = None,
    size_limit: int | None = None,
) -> bool:
    """
    See `SuccessionDiagram.expand_dfs` for documentation.
    """

    if node_id is None:
        node_id = sd.root()

    seen: set[int] = set()
    seen.add(node_id)

    stack: list[tuple[int, list[int] | None]] = [(node_id, None)]

    result_is_complete = True

    while len(stack) > 0:
        (node, successors) = stack.pop()
        if successors is None:
            # Only allow successor computation if size limit hasn't been exceeded.
            if (size_limit is not None) and (len(sd) >= size_limit):
                # Size limit reached.
                return False

            successors = sd.node_successors(node, compute=True)
            successors = sorted(successors, reverse=True)  # For determinism!
            # (reversed because we explore the list from the back)

        # Remove all immediate successors that are already visited.
        while len(successors) > 0 and successors[-1] in seen:
            successors.pop()

        # This node is done and we don't have to push anything onto the stack.
        if len(successors) == 0:
            continue

        if (dfs_stack_limit is not None) and (len(stack) >= dfs_stack_limit):
            # We cannot push any successor nodes because it would exceed
            # the stack limit. As such, we can just continue with the next
            # item on the stack. however, we must remember that we skipped
            # some nodes and the result is thus incomplete.
            result_is_complete = False
            continue

        s = successors.pop()
        seen.add(s)
        # Push the node back with the remaining successors.
        stack.append((node, successors))
        # Push the successor onto the stack.
        stack.append((s, None))

    return result_is_complete



from biobalm import control as control  # noqa: F401
from biobalm import drivers as drivers  # noqa: F401
from biobalm import types as types  # noqa: F401
from biobalm.succession_diagram import (  # noqa: F401
    SuccessionDiagram as SuccessionDiagram,
)

__version__ = "0.1.1"



from __future__ import annotations

import itertools as it
from typing import TYPE_CHECKING, cast

from biobalm.types import BooleanSpace
from biobalm.interaction_graph_utils import source_nodes

if TYPE_CHECKING:
    from biobalm.succession_diagram import SuccessionDiagram


def expand_source_blocks(
    sd: SuccessionDiagram,
    check_maa: bool = True,
    size_limit: int | None = None,
    optimize_source_nodes: bool = True,
    check_maa_exact: bool = False,
) -> bool:
    """
    Base correctness assumptions:

     - Expanding two minimal blocks is always independent. Every minimal trap space is contained in one of the maximal trap
       spaces of a single block. Furthermore, expanding a minimal block cannot intrfere with the expansion of the other
       stable motifs.
     - Any path in the "inner" succession diagram based on a single block can be reproduced in the larger succession diagram,
       meaning absence of attractors in the inner diagram proves absence of attractors in the outer diagram.
    """

    if sd.config["debug"]:
        print(f"Start SD expansion using block decomposition on {sd.network}.")

    root = sd.root()

    current_level: set[int] = set([root])
    next_level: set[int] = set()

    bfs_depth = 0

    while len(current_level) > 0:
        bfs_depth += 1
        if sd.config["debug"]:
            total_fixed_vars = 0
            for node in current_level:
                total_fixed_vars += len(sd.node_data(node)["space"])
            print(
                f" > Start block expansion of a BFS level {bfs_depth} with {len(current_level)} node(s). Average number of fixed variables: {total_fixed_vars / len(current_level)}/{sd.network.variable_count()}"
            )

        for node in sorted(current_level):  # Sorted for determinism
            if sd.node_data(node)["expanded"]:
                # We re-discovered a previously expanded node.
                continue

            # Only continue if the succession diagram isn't too large.
            if (size_limit is not None) and (len(sd) >= size_limit):
                # Size limit reached.
                return False

            node_bn = sd.node_percolated_network(node, compute=True)
            node_space = sd.node_data(node)["space"]

            # Check for source nodes in the percolated network. We can "fast-forward" these the
            # same way we would skip them in the root node.
            sources = source_nodes(node_bn)
            if len(sources) != 0 and optimize_source_nodes:
                if sd.config["debug"]:
                    print(
                        f" > Found {len(sources)} source nodes in node {node}. Fast-forwarding instead of expansion."
                    )

                # Check if size limits are exceeded.
                expected_size = len(sd) + (2 ** len(sources))
                if expected_size > sd.config["max_motifs_per_node"]:
                    raise RuntimeError(
                        f"Exceeded the maximum amount of stable motifs per node ({sd.config['max_motifs_per_node']}; see `SuccessionDiagramConfiguration.max_motifs_per_node`)."
                    )
                elif size_limit is not None and expected_size > size_limit:
                    # Cannot expand. Size limit would be exceeded.
                    return False
                else:
                    if sd.config["debug"]:
                        print(
                            f" > Fast-forwarding {len(sources)} source nodes into {2 ** len(sources)} SD nodes."
                        )

                bin_values_iter = it.product(range(2), repeat=len(sources))
                for bin_values in bin_values_iter:
                    valuation = cast(BooleanSpace, dict(zip(sources, bin_values)))
                    sub_space = node_space | valuation

                    next_level.add(sd._ensure_node(node, sub_space))  # type: ignore

                # This makes the node artificially "expanded". Also, there
                # can be no attractors here because we are just fixing the source nodes.
                sd.node_data(node)["expanded"] = True
                sd.node_data(node)["attractor_seeds"] = []
                sd.node_data(node)["attractor_sets"] = []

                # This node is done, more work will be done at the next level.
                continue

            # Compute successors as if this was a normal expansion procedure.
            successors = sd.node_successors(node, compute=True)
            # Sort successors to avoid non-determinism.
            successors = sorted(successors)

            if len(successors) == 0:
                # This is a minimal trap space.
                continue

            if len(successors) == 1 and not check_maa:
                # This space is not minimal, but there is no "choice" to make here.
                # We can just continue with that single choice.
                #
                # However, if we are checking for MAAs, we continue as normal, because
                # we need to compute most of the non-trivial results anyway.

                next_level.add(successors[0])
                continue

            # Split successors into "blocks" based on the regulatory component
            # of the variables fixed by the stable motif.

            # Maps a "block" (bwd-closed set of variables) to a list of node IDs (successor nodes).
            blocks: list[tuple[set[str], list[int]]] = []
            for s in successors:
                motif = sd.edge_stable_motif(node, s, reduced=True)
                motif_block = node_bn.backward_reachable(list(motif.keys()))
                motif_block_names = {node_bn.get_variable_name(v) for v in motif_block}

                found = False
                for block, nodes in blocks:
                    if block == motif_block_names:
                        found = True
                        nodes.append(s)
                        break
                if not found:
                    blocks.append((motif_block_names, [s]))

            if sd.config["debug"]:
                print(
                    f" > [{node}] Computed blocks: {[(len(k), len(v)) for (k, v) in blocks]}"
                )

            # Now remove all non-minimal blocks (minimal in terms of set inclusion).
            # The reason why we are removing those is that they are not independent
            # on the minimal blocks.

            minimal_blocks: list[tuple[set[str], list[int]]] = []
            if len(blocks) > 1:
                for block, nodes in blocks:
                    is_minimal = True
                    for b2, _ in blocks:
                        if b2 < block:
                            is_minimal = False
                            break
                    if is_minimal:
                        minimal_blocks.append((block, nodes))
            else:
                minimal_blocks = blocks

            # Sort minimal blocks by the number of successor nodes (we want to pick
            # the one that leads leads to the minimal SD expansion).
            minimal_blocks = sorted(minimal_blocks, key=lambda x: len(x[1]))

            if sd.config["debug"]:
                print(
                    f" > [{node}] Minimal blocks: {[(len(k), len(v)) for (k, v) in minimal_blocks]}"
                )

            if not check_maa:
                # We will expand all nodes that are in the smallest block.
                to_expand = minimal_blocks[0][1]

                if sd.config["debug"]:
                    print(f" > [{node}] Final block ({len(to_expand)}): {to_expand}")

                next_level = next_level | set(to_expand)
            else:
                # Here, we want to find the smallest block without any MAAs and choose it.
                # If such block does not exist, we expand the whole node, because the MAAs
                # can be either in this node, or in any of the child nodes. If a clean block
                # is found, we know that it is safe to expand it without "missing" any MAAs.
                clean_block_found = False
                for block, block_nodes in minimal_blocks:
                    block_sd = sd.component_subdiagram(list(block), node)

                    # Instead of expanding the inner succession diagram "normally", we can
                    # copy the successors/stable motifs that we already know. This doesn't
                    # make too much of a difference unless the maximal trap space computation
                    # is complex, but that is sometimes true, especially for large networks.
                    for succ_id in block_nodes:
                        succ_motif = sd.edge_stable_motif(node, succ_id, reduced=True)
                        block_sd._ensure_node(block_sd.root(), succ_motif)  # type: ignore
                    block_sd.node_data(block_sd.root())["expanded"] = True

                    # We could also consider using `seeds` instead of `candidates` here. Ultimately, this
                    # matters very rarely. The reasoning for why we use `candidates` is that we can (almost)
                    # always guarantee that the expansion is going to finish, even if some nodes do not
                    # have their MAAs eliminated. As such, one can try to use other techniques to disprove
                    # MAAs in the problematic nodes while using the nice properties of the expansion to
                    # still disprove MAAs in the remaining nodes. If we used `seeds`, the expansion could
                    # just get stuck on this node and the "partial" results wouldn't be usable.
                    is_clean = False
                    try:
                        if check_maa_exact:
                            block_sd_candidates = block_sd.node_attractor_seeds(
                                block_sd.root(), compute=True, symbolic_fallback=True
                            )
                        else:
                            block_sd_candidates = block_sd.node_attractor_candidates(
                                block_sd.root(), compute=True
                            )
                        is_clean = len(block_sd_candidates) == 0
                    except RuntimeError:
                        is_clean = False

                    if is_clean:
                        if sd.config["debug"]:
                            print(
                                f" > [{node}] Found clean block with no MAAs ({len(block_nodes)}): {block_nodes}"
                            )
                        clean_block_found = True
                        next_level = next_level | set(block_nodes)
                        sd.node_data(node)["attractor_seeds"] = []
                        sd.node_data(node)["attractor_sets"] = []
                        break
                    else:
                        if sd.config["debug"]:
                            print(
                                f"[{node}] > Found MAA candidates in a block (or failed candidate search). Delaying expansion."
                            )
                if not clean_block_found:
                    # If all blocks have MAAs, we expand all successors.
                    if sd.config["debug"]:
                        print(
                            f" > [{node}] No clean block found. Expanding all {len(successors)} successors."
                        )
                    next_level = next_level | set(successors)

        current_level = next_level
        next_level = set()

    if sd.config["debug"]:
        print(f" > Block expansion terminated with {len(sd)} node(s).")

    return True



"""
Utilities for translating Boolean networks into Petri nets.

Implements the translation from a `BooleanNetwork` object into a Petri net that
can be processed by Trappist (for finding trap spaces). The Petri net is
represented as a `networkx.DiGraph`, with nodes having either a `kind=place` or
a `kind=transition` attribute.

The variable names in the network have to be "sanitized" before translation. In
particular, this means they can't use any special characters beyond "_". In the
resulting Petri net, we then use a "b0_*" and "b1_*" prefix to distinguish the
"zero" and "one" places created for each variable. This is also important for
`clingo`, as we have to guarantee that in our logic program, symbols start with
a lowercase letter.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Generator

    from biodivine_aeon import Bdd, BooleanNetwork

    from biobalm.types import BooleanSpace

import copy
import re
from typing import cast

from biodivine_aeon import (
    BddPartialValuation,
    BddVariableSet,
    BoolType,
    SymbolicContext,
)
from networkx import DiGraph  # type: ignore

# Enables statistics logging.
DEBUG = False
"""Enables debug logging to stdout."""


def sanitize_network_names(network: BooleanNetwork, check_only: bool = False):
    """
    Rename variables in a network so that they can be safely used in Trappist.

    Verifies that all network variable names contain only alphanumeric
    characters and underscores. If this is not the case, attempts to rename the
    variables to make them compliant. Returns a *copy* of the original network
    that only uses sanitized names.

    Note that AEON should already prune away most of the special characters when
    parsing models, but it still allows names like `Gene_{Subscript}` (i.e.
    curly brackets) which we have to sanitize here.

    If `check_only=True` is specified, no renaming takes place and the function
    fails with a `RuntimeError` instead.

    Parameters
    ----------
    network : BooleanNetwork
        The network to sanitize.
    check_only : bool
        If `True`, no renaming takes place and the function fails with a
        `RuntimeError` if the network contains invalid variable names.

    Returns
    -------
    BooleanNetwork
        A copy of the original network with sanitized variable names.
    """
    network = copy.copy(network)
    for var in network.variables():
        name = network.get_variable_name(var)
        if not re.match("^[a-zA-Z0-9_]+$", name):
            if check_only:
                raise RuntimeError(f"Found unsanitized variable: `{name}`.")
            # Replace all invalid characters with an underscore
            new_name = re.sub("[^a-zA-Z0-9_]", "_", name)
            while True:
                try:
                    network.set_variable_name(var, new_name)
                    break
                except Exception:
                    # Name clash happened. Try to resolve this by adding an extra underscore to
                    # the variable name. In theory, this can repeat until a unique name is found.
                    new_name = "_" + new_name

    return network


def variable_to_place(variable: str, positive: bool) -> str:
    """
    Generate a Petri net place name from a network variable name.

    Parameters
    ----------
    variable : str
        The name of the network variable.
    positive : bool
        `True` if the place corresponding to the variable should be positive,
        `False` if it shoudl be negative.

    Returns
    -------
    str
        The name of the corresponding Petri net place.
    """
    if positive:
        return f"b1_{variable}"
    else:
        return f"b0_{variable}"


def place_to_variable(place: str) -> tuple[str, bool]:
    """
    Extract the variable name and state from a Petri net place name.

    Parameters
    ----------
    place : str
        The name of the Petri net place.

    Returns
    -------
    tuple[str, bool]
        The name of the variable, and whether the variable is positive or negative.
    """
    if place.startswith("b1_"):
        return (place[3:], True)
    elif place.startswith("b0_"):
        return (place[3:], False)
    else:
        raise Exception(f"Invalid place name: `{place}`.")


def extract_variable_names(encoded_network: DiGraph) -> list[str]:
    """
    Extract the variable names from a Petri net encoded Boolean network.

    The variables are  sorted lexicographically, since the original BN ordering
    is not preserved by the Petri net. However, BNs order variables
    lexicographically by default, so unless the Petri net was created from a
    custom BN (i.e. not from a normal model file), the ordering should be the
    same.

    Parameters
    ----------
    encoded_network : DiGraph
        The Petri net encoded Boolean network.

    Returns
    -------
    list[str]
        The list of variable names.
    """
    variables: list[str] = []
    for node in encoded_network.nodes():  # type: ignore
        node = str(node)  # type: ignore
        if node.startswith("b0_"):
            variables.append(place_to_variable(node)[0])

    return sorted(variables)


def extract_source_variables(encoded_network: DiGraph) -> list[str]:
    """
    List variable names that represent source nodes of the encoded network.

    Source nodes are those nodes with an identity update function.

    Parameters
    ----------
    encoded_network : DiGraph
        The Petri net encoded Boolean network.

    Returns
    -------
    list[str]
        The list of source variable names.
    """
    variables = extract_variable_names(encoded_network)
    source_set = set(variables)
    for _, change_var in encoded_network.nodes(data="change"):  # type: ignore
        if change_var in source_set:
            source_set.remove(change_var)  # type: ignore[reportUnknownArgumentType] # noqa
    source_nodes: list[str] = sorted(source_set)
    return source_nodes


def restrict_petrinet_to_subspace(
    petri_net: DiGraph,
    sub_space: BooleanSpace,
) -> DiGraph:
    """
    Create a copy of a Petri net restricted to a sub-space.

    Creates a copy of the given Petri net, but with the variables given in
    `sub_space` fixed to their respective values.

    Note that this completely eliminates the constant variables from the Petri
    net, but it does not perform any further constant propagation or
    percolation. Variables that are fixed in the `sub_space` but do not exist in
    the Petri net are ignored.

    The `sub_space` can contain variables that do not appear
    in the `petri_net`. Such variables are simply ignored.

    Parameters
    ----------
    petri_net : DiGraph
        The Petri net to restrict.
    sub_space : BooleanSpace
        The sub-space to restrict the Petri net to.

    Returns
    -------
    DiGraph
        The restricted Petri net.
    """
    result = copy.deepcopy(petri_net)
    for var, value in sub_space.items():
        # The idea is that we want to *remove* the place that corresponds to the fixed
        # value (it's effect on transitions is assumed to be fulfilled). Then, we remove
        # all transitions that depend on the inverse of the fixed value, as these can
        # never be satisfied. Then, we also remove inverse place.
        fixed_place = variable_to_place(var, bool(value))
        inverse_place = variable_to_place(var, not bool(value))

        if fixed_place not in result.nodes or inverse_place not in result.nodes:
            # These nodes were already removed.
            continue

        # First, remove all transitions that modify the fixed variable.
        # These are removed regardless of the actual value.

        to_delete: set[str] = set()

        # Transitions that put marker into one of the place values, but do not take it.
        for tr in result.predecessors(fixed_place):  # type: ignore
            if not result.has_edge(fixed_place, tr):  # type: ignore
                to_delete.add(cast(str, tr))
        for tr in result.predecessors(inverse_place):  # type: ignore
            if not result.has_edge(inverse_place, tr):  # type: ignore
                to_delete.add(cast(str, tr))

        # Transitions that depend on the value of the inverse place
        for tr in result.successors(inverse_place):  # type: ignore
            to_delete.add(cast(str, tr))

        for tr in to_delete:
            result.remove_node(tr)  # type: ignore

        result.remove_node(fixed_place)  # type: ignore
        result.remove_node(inverse_place)  # type: ignore
    return result


def network_to_petrinet(
    network: BooleanNetwork, symbolic_context: SymbolicContext | None = None
) -> DiGraph:
    """
    Convert a Boolean network to a Petri net.

    Converts a `biodivine_aeon.BooleanNetwork` to a `DiGraph` representing a Petri net encoding
    of the original network. For details about the encoding, see module
    description.

    Note that the given network needs to have "sanitized" names, otherwise the
    method will fail (see `sanitize_network_names` in this module).

    The operation uses translation through `biodivine_aeon.Bdd` to generate a
    disjunctive normal form of the network's update functions. This is
    facilitated by `biodivine_aeon.SymbolicContext`. If a suitable context already
    exists, it can be provided as the second argument. Otherwise it will be
    created.

    Parameters
    ----------
    network : BooleanNetwork
        The network to convert.
    symbolic_context : SymbolicContext | None
        The context used for the symbolic conversion, as an
        `biodivine_aeon.SymbolicContext` object. This is a mapping from the
        network nodes to BDD variables that preserves variable ordering in BDDs.
        If not given, a new one will be created from the network.

    Returns
    -------
    DiGraph
        The Petri net encoding of the given network.
    """
    # Assert that all network names are already sanitized.
    sanitize_network_names(network, check_only=True)

    assert (
        network.explicit_parameter_count() == 0
    ), f"Parametrized networks are not supported. Found parameters: {network.explicit_parameter_names()}."

    # Implicit parameters with no regulators are allowed, since they just reprtesent free inputs
    # and are explicitly handled by the succession diagram.
    non_input_implicit = [
        v for v in network.implicit_parameters() if len(network.predecessors(v)) > 0
    ]
    if len(non_input_implicit) > 0:
        names = [network.get_variable_name(x) for x in non_input_implicit]
        raise AssertionError(
            f"Parametrized networks are not supported. Found implicit parameters: {names}."
        )

    if symbolic_context is None:
        symbolic_context = SymbolicContext(network)

    pn = DiGraph()

    # Create a positive and a negative place for each variable.
    places: dict[str, tuple[str, str]] = {}
    for name in network.variable_names():
        p_name = variable_to_place(name, positive=True)
        n_name = variable_to_place(name, positive=False)
        pn.add_node(p_name, kind="place")  # type: ignore[reportUnknownMemberType]
        pn.add_node(n_name, kind="place")  # type: ignore[reportUnknownMemberType]
        places[name] = (n_name, p_name)

    # Create PN transitions for implicants of every BN transition.
    for var in network.variables():
        var_name = network.get_variable_name(var)
        update_function = network.get_update_function(var)
        if update_function is None:
            # This variable is a constant input with a free update function.
            assert len(network.predecessors(var)) == 0
            continue

        function_bdd = symbolic_context.mk_update_function(update_function)
        var_bdd = symbolic_context.mk_network_variable(var)

        p_bdd = function_bdd.l_and(var_bdd.l_not())
        n_bdd = function_bdd.l_not().l_and(var_bdd)

        if DEBUG:
            print(f"Start translation for `{var_name}`: {len(p_bdd)} | {len(n_bdd)}")

        # Add 0->1 edges.
        _create_transitions(
            pn, symbolic_context.bdd_variable_set(), places, var_name, p_bdd, go_up=True
        )
        # Add 1-> 0 edges.
        _create_transitions(
            pn,
            symbolic_context.bdd_variable_set(),
            places,
            var_name,
            n_bdd,
            go_up=False,
        )

    return pn


def optimized_recursive_dnf_generator(
    bdd: Bdd,
) -> Generator[BddPartialValuation, None, None]:
    """
    Yields a generator of `BddPartialValuation` objects, similar to
    `Bdd.clause_iterator` in AEON, but uses a recursive optimization strategy
    to return a smaller result than the default method `Bdd` clause sequence.
    Note that this is still not the "optimal" DNF, but is often close enough.

    This is technically slower for BDDs that already have a small clause count,
    but can be much better in the long-term when the clause count is high.
    """

    # Some performance notes:
    #  - We could cache the best restriced BDDs, but usually this does not help much.
    #  - We could try optimizing based on clause count instead of BDD size, but this will
    #    need a new method in `lib-bdd` to compute clause count efficiently. (Now
    #    we can only do it by walking the iterator.)
    #  - The initial BDD size/ordering still do matter: starting with a smaller but equivalent
    #    BDD can help achieve smaller DNF size.

    if bdd.is_false():
        return
    if bdd.is_true():
        yield BddPartialValuation(bdd.__ctx__(), cast(dict[str, BoolType], {}))
        return

    support = sorted(bdd.support_set())
    best_var = support[0]
    best_size = 10 * len(bdd)
    for var in support:
        t = bdd.r_restrict({var: True})
        f = bdd.r_restrict({var: False})
        t_size = len(t)
        f_size = len(f)
        if t_size + f_size < best_size:
            best_size = t_size + f_size
            best_var = var

    for t_val in optimized_recursive_dnf_generator(bdd.r_restrict({best_var: True})):
        t_val[best_var] = True
        yield t_val

    for f_val in optimized_recursive_dnf_generator(bdd.r_restrict({best_var: False})):
        f_val[best_var] = False
        yield f_val


def _create_transitions(
    pn: DiGraph,
    ctx: BddVariableSet,
    places: dict[str, tuple[str, str]],
    var_name: str,
    implicant_bdd: Bdd,
    go_up: bool,
):
    """
    Just a helper method that creates PN transitions from BDDs representing
    positive/negative BN transitions.
    """
    dir_str = "up" if go_up else "down"
    total = 0
    for t_id, implicant in enumerate(optimized_recursive_dnf_generator(implicant_bdd)):
        total += 1
        t_name = f"tr_{var_name}_{dir_str}_{t_id + 1}"
        pn.add_node(  # type: ignore[reportUnknownMemberType]
            t_name,
            kind="transition",  # type: ignore[reportUnknownMemberType]
            change=var_name,
            direction=dir_str,
        )
        # The transition moves a token either from "zero place" to the
        # "one place", or vice versa.
        if go_up:
            pn.add_edge(places[var_name][0], t_name)  # type: ignore[reportUnknownMemberType] # noqa
            pn.add_edge(t_name, places[var_name][1])  # type: ignore[reportUnknownMemberType] # noqa
        else:
            pn.add_edge(places[var_name][1], t_name)  # type: ignore[reportUnknownMemberType] # noqa
            pn.add_edge(t_name, places[var_name][0])  # type: ignore[reportUnknownMemberType] # noqa
        for variable, value in implicant.items():
            variable_str = ctx.get_variable_name(
                variable
            )  # Convert from BDD variable to name.
            if variable_str == var_name:
                continue
            # For the remaining variables, we simply check if the required
            # token is present in the corresponding place.
            pn.add_edge(places[variable_str][value], t_name)  # type: ignore[reportUnknownMemberType] # noqa
            pn.add_edge(t_name, places[variable_str][value])  # type: ignore[reportUnknownMemberType] # noqa
    if DEBUG:
        print(f"  >> Generated {total} total PN transitions.")



"""
A module which provides functions for processing the stable motif driver nodes of a Boolean network.
"""

from __future__ import annotations

from biodivine_aeon import AsynchronousGraph, BooleanNetwork

from biobalm.space_utils import percolate_space_strict
from biobalm.types import BooleanSpace


def find_single_node_LDOIs(
    network: AsynchronousGraph | BooleanNetwork,
) -> dict[tuple[str, int], BooleanSpace]:
    """
    Finds the the logical domain of influence for every single node state.

    The logical domain of influence is equivalent to the set of node states that
    become fixed after percolation. See
    :func:`percolate_space<biobalm.space_utils.percolate_space>` for details.

    This operation requires the symbolic update functions provided by an
    `AsynchronousGraph`. If you provide a `BooleanNetwork`, the
    `AsynchronousGraph` will be created automatically, but this can incur
    additional overhead.

    Parameters
    ----------
    network : AsynchronousGraph | BooleanNetwork
        The symbolic update functions stored as a `AsynchronousGraph` or
        `BooleanNetwork` object from the `biodivine_aeon` library.

    Returns
    -------
    dict[tuple[str, int], BooleanSpace]
        The LDOI for every one-node state. The keys are `(variable, value)`
        tuples, where `variable` is the name of the node that is fixed in the
        state `value` (`0` or `1`). The keys are
        :class:`BooleanSpace<biobalm.types.BooleanSpace>` objects, which are dictionaries
        of node values describing the node states that become fixed as a result
        of percolation.

    Example
    _______
    >>> import biobalm
    >>> sd = biobalm.SuccessionDiagram.from_rules('A, A\\nB, A')
    >>> ldois = biobalm.drivers.find_single_node_LDOIs(sd.network)
    >>> for k in sorted(ldois):
    ...     print(f'{k} ==> {sorted(ldois[k].items())}')
    ...
    ('A', 0) ==> [('A', 0), ('B', 0)]
    ('A', 1) ==> [('A', 1), ('B', 1)]
    ('B', 0) ==> []
    ('B', 1) ==> []
    """
    if isinstance(network, BooleanNetwork):
        network = AsynchronousGraph(network)

    LDOIs: dict[tuple[str, int], BooleanSpace] = {}
    for var in network.network_variable_names():
        fn_bdd = network.mk_update_function(var)
        if fn_bdd.is_true() or fn_bdd.is_false():
            # Skip constant nodes.
            continue

        LDOIs[(var, 0)] = percolate_space_strict(network, {var: 0})
        LDOIs[(var, 1)] = percolate_space_strict(network, {var: 1})

    return LDOIs


def find_single_drivers(
    target_subspace: BooleanSpace,
    network: AsynchronousGraph | BooleanNetwork,
    LDOIs: dict[tuple[str, int], BooleanSpace] | None = None,
) -> set[tuple[str, int]]:
    """
    Find all one-node drivers for a target subspace.

    This operation requires the symbolic update functions provided by an
    `AsynchronousGraph`. If you provide a `BooleanNetwork`, the
    `AsynchronousGraph` will be created automatically, but this can incur
    additional overhead.

    Parameters
    ----------
    target_subspace : BooleanSpace
        A :class:`BooleanSpace<biobalm.types.BooleanSpace>` object describing the
        target trap space.
    network : AsynchronousGraph | BooleanNetwork
       The symbolic update functions stored as a `AsynchronousGraph` or
        `BooleanNetwork` object from the `biodivine_aeon` library.
    LDOIs : dict[tuple[str, int], BooleanSpace] | None, optional
        A dictionary encoding the LDOI for every one-node state to be
        considered. The keys should be `(variable, value)` tuples, where
        `variable` is the name of the node that is fixed in the state `value`
        (`0` or `1`). The keys are
        :class:`BooleanSpace<biobalm.types.BooleanSpace>` objects, which are
        dictionaries of node values describing the node states that become fixed
        as a result of percolation. If not provided, the LDOI will be computed
        automatically for each node state using :func:`find_single_node_LDOIs`.


    Returns
    -------
    set[tuple[str, int]]
        A set of driver node states that result in the target subspace, with
        each node state represented as a node name and value (`0` or `1`).

    Example
    -------
    >>> import biobalm
    >>> sd = biobalm.SuccessionDiagram.from_rules('A, A\\nB, A')
    >>> drivers = biobalm.drivers.find_single_drivers({'B': 0}, sd.network)
    >>> sorted(list(drivers))
    [('A', 0), ('B', 0)]
    """
    if isinstance(network, BooleanNetwork):
        network = AsynchronousGraph(network)

    if LDOIs is None:
        LDOIs = find_single_node_LDOIs(network)

    drivers: set[tuple[str, int]] = set()
    for fix, LDOI in LDOIs.items():
        if target_subspace.items() <= (LDOI.items() | {fix}):
            drivers.add(fix)

    return drivers
